from django import forms
from django.utils.translation import pgettext_lazy

from .base import SettingsForm


class ThreadsSettingsForm(SettingsForm):
    settings = [
        "attachment_403_image",
        "attachment_404_image",
        "daily_post_limit",
        "hourly_post_limit",
        "post_attachments_limit",
        "post_length_max",
        "post_length_min",
        "readtracker_cutoff",
        "thread_title_length_max",
        "thread_title_length_min",
        "unused_attachments_lifetime",
        "threads_per_page",
        "posts_per_page",
        "posts_per_page_orphans",
        "events_per_page",
    ]

    daily_post_limit = forms.IntegerField(
        label=pgettext_lazy("admin threads settings form", "Daily post limit per user"),
        help_text=pgettext_lazy(
            "admin threads settings form",
            "Daily limit of posts that may be posted by single user. Fail-safe for situations when forum is flooded by spam bots. Change to 0 to remove the limit.",
        ),
        min_value=0,
    )
    hourly_post_limit = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form", "Hourly post limit per user"
        ),
        help_text=pgettext_lazy(
            "admin threads settings form",
            "Hourly limit of posts that may be posted by single user. Fail-safe for situations when forum is flooded by spam bots. Change to 0 to remove the limit.",
        ),
        min_value=0,
    )
    post_attachments_limit = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form", "Maximum number of attachments per post"
        ),
        min_value=1,
    )
    post_length_max = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form", "Maximum allowed post length"
        ),
        min_value=0,
    )
    post_length_min = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form", "Minimum required post length"
        ),
        min_value=1,
    )
    thread_title_length_max = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form", "Maximum allowed thread title length"
        ),
        min_value=2,
        max_value=255,
    )
    thread_title_length_min = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form", "Minimum required thread title length"
        ),
        min_value=2,
        max_value=255,
    )
    unused_attachments_lifetime = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form", "Unused attachments lifetime"
        ),
        help_text=pgettext_lazy(
            "admin threads settings form",
            "Period of time (in hours) after which user-uploaded files that weren't attached to any post are deleted from disk.",
        ),
        min_value=1,
    )

    readtracker_cutoff = forms.IntegerField(
        label=pgettext_lazy("admin threads settings form", "Read-tracker cutoff"),
        help_text=pgettext_lazy(
            "admin threads settings form",
            "Controls amount of data used by read-tracking system. All content older than number of days specified in this setting is considered old and read, even if the opposite is true for the user. Active forums can try lowering this value while less active ones may wish to increase it instead.",
        ),
        min_value=1,
    )

    threads_per_page = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form",
            "Number of threads displayed on a single page",
        ),
        min_value=10,
    )

    posts_per_page = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form", "Number of posts displayed on a single page"
        ),
        min_value=5,
    )
    posts_per_page_orphans = forms.IntegerField(
        label=pgettext_lazy("admin threads settings form", "Maximum orphans"),
        help_text=pgettext_lazy(
            "admin threads settings form",
            "This setting prevents situations when the last page of a thread contains very few items. If number of posts to be displayed on the last page is less or equal to number specified in this setting, those posts will instead be appended to the previous page, reducing number of thread's pages.",
        ),
        min_value=0,
    )
    events_per_page = forms.IntegerField(
        label=pgettext_lazy(
            "admin threads settings form",
            "Maximum number of events displayed on a single page",
        ),
        min_value=5,
    )

    attachment_403_image = forms.ImageField(
        label=pgettext_lazy("admin threads settings form", "Permission denied"),
        help_text=pgettext_lazy(
            "admin threads settings form",
            "Attachments proxy will display this image in place of default one when user tries to access attachment they have no permission to see.",
        ),
        required=False,
    )
    attachment_403_image_delete = forms.BooleanField(
        label=pgettext_lazy(
            "admin threads settings form", "Delete custom permission denied image"
        ),
        required=False,
    )
    attachment_404_image = forms.ImageField(
        label=pgettext_lazy("admin threads settings form", "Not found"),
        help_text=pgettext_lazy(
            "admin threads settings form",
            "Attachments proxy will display this image in place of default one when user tries to access attachment that doesn't exist.",
        ),
        required=False,
    )
    attachment_404_image_delete = forms.BooleanField(
        label=pgettext_lazy(
            "admin threads settings form", "Delete custom not found image"
        ),
        required=False,
    )

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("posts_per_page_orphans") > cleaned_data.get(
            "posts_per_page"
        ):
            self.add_error(
                "posts_per_page_orphans",
                pgettext_lazy(
                    "admin threads settings form",
                    "This value must be lower than number of posts per page.",
                ),
            )
        return cleaned_data



from ....test import assert_contains


def test_icons_admin_view_displays(admin_client, admin_link):
    response = admin_client.get(admin_link)
    assert response.status_code == 200


def test_set_favicon_displays(admin_client, admin_link, favicon):
    response = admin_client.get(admin_link)
    assert_contains(response, favicon.image.url)


def test_set_favicon_32_displays(admin_client, admin_link, favicon_32):
    response = admin_client.get(admin_link)
    assert_contains(response, favicon_32.image.url)


def test_set_favicon_16_displays(admin_client, admin_link, favicon_16):
    response = admin_client.get(admin_link)
    assert_contains(response, favicon_16.image.url)


def test_set_apple_touch_icon_displays(admin_client, admin_link, apple_touch_icon):
    response = admin_client.get(admin_link)
    assert_contains(response, apple_touch_icon.image.url)



import random
import time

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from faker import Factory

from ....categories.models import Category
from ....core.management.progressbar import show_progress
from ....threads.models import Thread
from ...threads import (
    get_fake_closed_thread,
    get_fake_hidden_thread,
    get_fake_thread,
    get_fake_unapproved_thread,
)

User = get_user_model()


class Command(BaseCommand):
    help = "Creates random threads for dev and testing purposes."

    def add_arguments(self, parser):
        parser.add_argument(
            "threads",
            help="number of threads to create",
            nargs="?",
            type=int,
            default=5,
        )

    def handle(
        self, *args, **options
    ):  # pylint: disable=too-many-locals, too-many-branches
        items_to_create = options["threads"]
        fake = Factory.create()

        categories = list(Category.objects.all_categories())

        message = "Creating %s fake threads...\n"
        self.stdout.write(message % items_to_create)

        created_threads = 0
        start_time = time.time()
        show_progress(self, created_threads, items_to_create)

        while created_threads < items_to_create:
            category = random.choice(categories)

            # 10% chance thread poster is anonymous
            if random.randint(0, 100) > 90:
                starter = None
            else:
                starter = User.objects.order_by("?").last()

            # There's 10% chance thread is closed
            if random.randint(0, 100) > 90:
                thread = get_fake_closed_thread(fake, category, starter)

            # There's further 5% chance thread is hidden
            elif random.randint(0, 100) > 95:
                if random.randint(0, 100) > 90:
                    hidden_by = None
                else:
                    hidden_by = User.objects.order_by("?").last()

                thread = get_fake_hidden_thread(fake, category, starter, hidden_by)

            # And further 5% chance thread is unapproved
            elif random.randint(0, 100) > 95:
                thread = get_fake_unapproved_thread(fake, category, starter)

            # Default, standard thread
            else:
                thread = get_fake_thread(fake, category, starter)

            thread.synchronize()
            thread.save()

            created_threads += 1
            show_progress(self, created_threads, items_to_create, start_time)

        pinned_threads = random.randint(0, int(created_threads * 0.025)) or 1
        self.stdout.write("\nPinning %s threads..." % pinned_threads)

        for _ in range(0, pinned_threads):
            thread = Thread.objects.order_by("?")[:1][0]
            if random.randint(0, 100) > 90:
                thread.weight = 2
            else:
                thread.weight = 1
            thread.save()

        for category in categories:
            category.synchronize()
            category.save()

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))
        message = "\nSuccessfully created %s fake threads in %s"
        self.stdout.write(message % (created_threads, total_humanized))



from django import forms
from django.utils.translation import pgettext_lazy

from ....admin.forms import YesNoSwitch
from .base import SettingsForm


class GeneralSettingsForm(SettingsForm):
    settings = [
        "forum_name",
        "forum_address",
        "index_header",
        "index_title",
        "index_message",
        "index_meta_description",
        "logo",
        "logo_small",
        "logo_text",
        "og_image",
        "og_image_avatar_on_profile",
        "og_image_avatar_on_thread",
        "forum_footnote",
        "email_footer",
        "show_admin_panel_link_in_ui",
    ]

    forum_name = forms.CharField(
        label=pgettext_lazy("admin general settings form", "Forum name"),
        min_length=2,
        max_length=255,
    )
    forum_address = forms.URLField(
        label=pgettext_lazy("admin general settings form", "Forum address"),
        max_length=255,
    )

    index_title = forms.CharField(
        label=pgettext_lazy("admin general settings form", "Page title"),
        max_length=255,
        required=False,
    )
    index_meta_description = forms.CharField(
        label=pgettext_lazy("admin general settings form", "Meta Description"),
        help_text=pgettext_lazy(
            "admin general settings form",
            "Short description of your forum that search and social sites may display next to link to your forum's index.",
        ),
        max_length=255,
        required=False,
    )
    index_header = forms.CharField(
        label=pgettext_lazy("admin general settings form", "Header text"),
        help_text=pgettext_lazy(
            "admin general settings form",
            "This text will be displayed in page header on forum index.",
        ),
        max_length=255,
        required=False,
    )
    index_message = forms.CharField(
        label=pgettext_lazy("admin general settings form", "Header message"),
        help_text=pgettext_lazy(
            "admin general settings form",
            "This message will be displayed in page header on forum index, under the header text.",
        ),
        max_length=2048,
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )

    logo = forms.ImageField(
        label=pgettext_lazy("admin general settings form", "Large logo"),
        help_text=pgettext_lazy(
            "admin general settings form",
            "Image that will be displayed in forum navbar instead of a small logo or text.",
        ),
        required=False,
    )
    logo_delete = forms.BooleanField(
        label=pgettext_lazy("admin general settings form", "Delete large logo image"),
        required=False,
    )
    logo_small = forms.ImageField(
        label=pgettext_lazy("admin general settings form", "Small logo"),
        help_text=pgettext_lazy(
            "admin general settings form",
            "Image that will be displayed in the forum navbar next to the logo text if a large logo was not uploaded.",
        ),
        required=False,
    )
    logo_small_delete = forms.BooleanField(
        label=pgettext_lazy("admin general settings form", "Delete small logo image"),
        required=False,
    )
    logo_text = forms.CharField(
        label=pgettext_lazy("admin general settings form", "Text logo"),
        help_text=pgettext_lazy(
            "admin general settings form",
            "Text displayed in forum navbar. If a small logo image was uploaded, this text will be displayed right next to it. If a large logo was uploaded, it will replace both the small logo and the text.",
        ),
        max_length=255,
        required=False,
    )

    og_image = forms.ImageField(
        label=pgettext_lazy("admin general settings form", "Image"),
        help_text=pgettext_lazy(
            "admin general settings form",
            "Custom image that will appear next to links to your forum posted on social sites. Facebook recommends that this image should be 1200 pixels wide and 630 pixels tall.",
        ),
        required=False,
    )
    og_image_delete = forms.BooleanField(
        label=pgettext_lazy("admin general settings form", "Delete image"),
        required=False,
    )
    og_image_avatar_on_profile = YesNoSwitch(
        label=pgettext_lazy(
            "admin general settings form", "Replace image with avatar on user profiles"
        )
    )
    og_image_avatar_on_thread = YesNoSwitch(
        label=pgettext_lazy(
            "admin general settings form", "Replace image with avatar on threads"
        )
    )

    forum_footnote = forms.CharField(
        label=pgettext_lazy("admin general settings form", "Forum footnote"),
        help_text=pgettext_lazy(
            "admin general settings form", "Short message displayed in forum footer."
        ),
        max_length=300,
        required=False,
    )
    email_footer = forms.CharField(
        label=pgettext_lazy("admin general settings form", "E-mails footer"),
        help_text=pgettext_lazy(
            "admin general settings form",
            "Optional short message included at the end of e-mails sent by forum.",
        ),
        max_length=255,
        required=False,
    )

    show_admin_panel_link_in_ui = YesNoSwitch(
        label=pgettext_lazy(
            "admin general settings form",
            "Display the link to the Admin Control Panel in the administrator's user menu",
        ),
        help_text=pgettext_lazy(
            "admin general settings form",
            "Hiding the link to the ACP from user menus reduces risk of malicious actors gaining access to admin session for admin users who are sharing their device with others or who are logging in to the site in public spaces.",
        ),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        address = self.request.build_absolute_uri("/")
        self["forum_address"].help_text = pgettext_lazy(
            "admin general settings form",
            'Misago uses this setting to build links in e-mails sent to site users. Address under which site is running appears to be "%(address)s".',
        ) % {"address": address}

    def clean_forum_address(self):
        return self.cleaned_data["forum_address"].lower()



from urllib.parse import urlparse

from django.conf import settings
from django.contrib import auth
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect
from django.urls import NoReverseMatch
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.translation import pgettext
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters


@sensitive_post_parameters()
@never_cache
@csrf_protect
def login(request):
    if request.settings.enable_oauth2_client:
        raise PermissionDenied(
            pgettext("login", "Please use %(provider)s to sign in.")
            % {"provider": request.settings.oauth2_provider}
        )

    if request.method == "POST":
        redirect_to = request.POST.get("redirect_to")
        if redirect_to:
            is_redirect_safe = url_has_allowed_host_and_scheme(
                url=redirect_to,
                allowed_hosts={request.get_host()},
                require_https=request.is_secure(),
            )
            if is_redirect_safe:
                redirect_to_path = urlparse(redirect_to).path
                if "?" not in redirect_to_path:
                    redirect_to_path = "%s?" % redirect_to_path
                else:
                    redirect_to_path = "%s&" % redirect_to_path
                redirect_to_path = "%sref=login" % redirect_to_path
                try:
                    return redirect(redirect_to_path)
                except NoReverseMatch:
                    pass

    return redirect(settings.LOGIN_REDIRECT_URL)


@never_cache
@csrf_protect
def logout(request):
    if request.method == "POST" and request.user.is_authenticated:
        auth.logout(request)
    return redirect(settings.LOGIN_REDIRECT_URL)



from .versions import get_cache_versions


def cache_versions_middleware(get_response):
    """Sets request.cache_versions attribute with dict of cache versions."""

    def middleware(request):
        request.cache_versions = get_cache_versions()
        return get_response(request)

    return middleware



"""
Changed credentials service

Stores new e-mail and password in cache
"""
from hashlib import sha256

from django.conf import settings
from django.utils.encoding import force_bytes


def store_new_credential(request, credential_type, credential_value):
    credential_key = "new_credential_%s" % credential_type
    token = _make_change_token(request.user, credential_type)

    request.session[credential_key] = {
        "user_pk": request.user.pk,
        "credential": credential_value,
        "token": token,
    }

    return token


def read_new_credential(request, credential_type, link_token):
    try:
        credential_key = "new_credential_%s" % credential_type
        new_credential = request.session.pop(credential_key)
    except KeyError:
        return None

    if new_credential["user_pk"] != request.user.pk:
        return None

    current_token = _make_change_token(request.user, credential_type)
    if link_token != current_token:
        return None
    if new_credential["token"] != current_token:
        return None

    return new_credential["credential"]


def _make_change_token(user, token_type):
    seeds = (
        user.pk,
        user.email,
        user.password,
        user.last_login.replace(microsecond=0, tzinfo=None),
        settings.SECRET_KEY,
        str(token_type),
    )

    return sha256(force_bytes("+".join([str(s) for s in seeds]))).hexdigest()



import pytest
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ....test import assert_has_error_message
from ... import THEME_CACHE
from ...models import Theme


@pytest.fixture
def activate_link(theme):
    return reverse("misago:admin:themes:activate", kwargs={"pk": theme.pk})


def test_active_theme_can_changed(admin_client, activate_link, theme):
    admin_client.post(activate_link)
    theme.refresh_from_db()
    assert theme.is_active


def test_default_theme_can_be_set_as_active_theme(admin_client, default_theme):
    activate_link = reverse(
        "misago:admin:themes:activate", kwargs={"pk": default_theme.pk}
    )
    admin_client.post(activate_link)
    default_theme.refresh_from_db()
    assert default_theme.is_active


def test_changing_active_theme_removes_active_status_from_previous_active_theme(
    admin_client, activate_link, theme
):
    admin_client.post(activate_link)
    # objets.get() will raise if more than one theme is active
    assert Theme.objects.get(is_active=True) == theme


def test_changing_active_theme_to_nonexisting_theme_sets_error_message(
    admin_client, nonexisting_theme
):
    activate_link = reverse(
        "misago:admin:themes:activate", kwargs={"pk": nonexisting_theme.pk}
    )
    response = admin_client.post(activate_link)
    assert_has_error_message(response)


def test_changing_active_theme_invalidates_themes_cache(admin_client, activate_link):
    with assert_invalidates_cache(THEME_CACHE):
        admin_client.post(activate_link)



class AdminBaseMixin:
    """
    Admin mixin abstraciton used for configuring admin CRUD views.

    Takes following attributes:

    model - Model instance
    root_link - name of link leading to root action (eg. list of all items
    templates_dir - directory with templates
    message_404 - string used in "requested item not found" messages
    """

    model = None
    root_link = None
    templates_dir = None
    message_404 = None

    def get_model(self):
        """basic method for retrieving Model, used in cases such as User model."""
        return self.model



from django import forms
from django.db.models import Q
from django.utils.translation import pgettext_lazy

from ...admin.forms import YesNoSwitch
from ..models import Agreement
from .utils import disable_agreement, set_agreement_as_active


class AgreementForm(forms.ModelForm):
    type = forms.ChoiceField(
        label=pgettext_lazy("admin agreement form", "Type"),
        choices=Agreement.TYPE_CHOICES,
    )
    title = forms.CharField(
        label=pgettext_lazy("admin agreement form", "Title"),
        help_text=pgettext_lazy(
            "admin agreement form",
            "Optional, leave empty for agreement to be named after its type.",
        ),
        required=False,
    )
    is_active = YesNoSwitch(
        label=pgettext_lazy("admin agreement form", "Active for its type"),
        help_text=pgettext_lazy(
            "admin agreement form",
            "If other agreement is already active for this type, it will be unset and replaced with this one. Misago will ask users who didn't accept this agreement to do so before allowing them to continue using the site.",
        ),
    )
    link = forms.URLField(
        label=pgettext_lazy("admin agreement form", "Link"),
        help_text=pgettext_lazy(
            "admin agreement form",
            "If your agreement is located on other page, enter here a link to it.",
        ),
        required=False,
    )
    text = forms.CharField(
        label=pgettext_lazy("admin agreement form", "Text"),
        help_text=pgettext_lazy(
            "admin agreement form",
            "You can use Markdown syntax for rich text elements.",
        ),
        widget=forms.Textarea,
        required=False,
    )

    class Meta:
        model = Agreement
        fields = ["type", "title", "link", "text", "is_active"]

    def clean(self):
        data = super().clean()

        if not data.get("link") and not data.get("text"):
            raise forms.ValidationError(
                pgettext_lazy(
                    "admin agreement form", "Please fill in agreement link or text."
                )
            )

        return data

    def save(self):
        agreement = super().save()
        if agreement.is_active:
            set_agreement_as_active(agreement)
        else:
            disable_agreement(agreement)
        Agreement.objects.invalidate_cache()
        return agreement


class FilterAgreementsForm(forms.Form):
    type = forms.ChoiceField(
        label=pgettext_lazy("admin agreement form", "Type"),
        required=False,
        choices=[("", pgettext_lazy("admin agreement type field choice", "All types"))]
        + Agreement.TYPE_CHOICES,
    )
    content = forms.CharField(
        label=pgettext_lazy("admin agreement form", "Content"), required=False
    )

    def filter_queryset(self, criteria, queryset):
        if criteria.get("type") is not None:
            queryset = queryset.filter(type=criteria["type"])

        if criteria.get("content"):
            search_title = Q(title__icontains=criteria["content"])
            search_text = Q(text__icontains=criteria["content"])
            queryset = queryset.filter(search_title | search_text)

        return queryset



from django.urls import path

from .views import notifications, notifications_read_all

urlpatterns = [
    path(
        "notifications/",
        notifications,
        name="notifications",
    ),
    path(
        "notifications/read-all/",
        notifications_read_all,
        name="notifications-read-all",
    ),
]



from django.contrib.auth import get_user_model
from django.core import mail
from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ...legal.models import Agreement
from ..models import Ban, Online
from ..test import UserTestCase

User = get_user_model()


class UserCreateTests(UserTestCase):
    """tests for new user registration (POST to /api/users/)"""

    def setUp(self):
        super().setUp()

        Agreement.objects.invalidate_cache()

        self.api_link = "/api/users/"

    def tearDown(self):
        Agreement.objects.invalidate_cache()

    def test_empty_request(self):
        """empty request errors with code 400"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "username": ["This field is required."],
                "email": ["This field is required."],
                "password": ["This field is required."],
            },
        )

    def test_invalid_data(self):
        """invalid request data errors with code 400"""
        response = self.client.post(
            self.api_link, "false", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "username": ["This field is required."],
                "email": ["This field is required."],
                "password": ["This field is required."],
            },
        )

    def test_authenticated_request(self):
        """authentiated user request errors with code 403"""
        self.login_user(self.get_authenticated_user())
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This action is not available to signed in users."},
        )

    @override_dynamic_settings(account_activation="closed")
    def test_registration_off_request(self):
        """registrations off request errors with code 403"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "New users registrations are currently closed."}
        )

    def test_registration_validates_ip_ban(self):
        """api validates ip ban"""
        Ban.objects.create(
            check_type=Ban.IP,
            banned_value="127.*",
            user_message="You can't register account like this.",
        )

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 403)

    def test_registration_validates_ip_registration_ban(self):
        """api validates ip registration-only ban"""
        Ban.objects.create(
            check_type=Ban.IP,
            banned_value="127.*",
            user_message="You can't register account like this.",
            registration_only=True,
        )

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"__all__": ["You can't register account like this."]}
        )

    def test_registration_validates_username(self):
        """api validates usernames"""
        user = self.get_authenticated_user()

        response = self.client.post(
            self.api_link,
            data={
                "username": user.username,
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"username": ["This username is not available."]},
        )

    def test_registration_validates_username_ban(self):
        """api validates username ban"""
        Ban.objects.create(
            banned_value="totally*",
            user_message="You can't register account like this.",
        )

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"username": ["You can't register account like this."]}
        )

    def test_registration_validates_username_registration_ban(self):
        """api validates username registration-only ban"""
        Ban.objects.create(
            banned_value="totally*",
            user_message="You can't register account like this.",
            registration_only=True,
        )

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"username": ["You can't register account like this."]}
        )

    def test_registration_validates_email(self):
        """api validates usernames"""
        user = self.get_authenticated_user()

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": user.email,
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"email": ["This e-mail address is not available."]},
        )

    def test_registration_validates_email_ban(self):
        """api validates email ban"""
        Ban.objects.create(
            check_type=Ban.EMAIL,
            banned_value="lorem*",
            user_message="You can't register account like this.",
        )

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"email": ["You can't register account like this."]}
        )

    def test_registration_validates_email_registration_ban(self):
        """api validates email registration-only ban"""
        Ban.objects.create(
            check_type=Ban.EMAIL,
            banned_value="lorem*",
            user_message="You can't register account like this.",
            registration_only=True,
        )

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"email": ["You can't register account like this."]}
        )

    def test_registration_requires_password(self):
        """api uses django's validate_password to validate registrations"""
        response = self.client.post(
            self.api_link,
            data={"username": "User", "email": "loremipsum@dolor.met", "password": ""},
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"password": ["This field is required."]})

    def test_registration_validates_password(self):
        """api uses django's validate_password to validate registrations"""
        response = self.client.post(
            self.api_link,
            data={
                "username": "User",
                "email": "l.o.r.e.m.i.p.s.u.m@gmail.com",
                "password": "123",
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "email": ["This e-mail is not allowed."],
                "password": [
                    (
                        "This password is too short. "
                        "It must contain at least 7 characters."
                    ),
                    "This password is entirely numeric.",
                ],
            },
        )

    def test_registration_validates_password_similiarity(self):
        """api uses validate_password to validate registrations"""
        response = self.client.post(
            self.api_link,
            data={
                "username": "BobBoberson",
                "email": "l.o.r.e.m.i.p.s.u.m@gmail.com",
                "password": "BobBoberson",
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "email": ["This e-mail is not allowed."],
                "password": ["The password is too similar to the username."],
            },
        )

    @override_dynamic_settings(
        captcha_type="qa", qa_question="Test", qa_answers="Lorem\nIpsum"
    )
    def test_registration_validates_captcha(self):
        """api validates captcha"""
        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
                "captcha": "dolor",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"captcha": ["Entered answer is incorrect."]})

        # valid captcha
        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
                "captcha": "ipSUM",
            },
        )

        self.assertEqual(response.status_code, 200)

    @override_dynamic_settings(
        captcha_type="qa", qa_question="", qa_answers="Lorem\n\nIpsum"
    )
    def test_qacaptcha_handles_empty_answers(self):
        """api validates captcha"""
        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
                "captcha": "",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"captcha": ["Entered answer is incorrect."]})

    def test_registration_check_agreement(self):
        """api checks agreement"""
        agreement = Agreement.objects.create(
            type=Agreement.TYPE_TOS, text="Lorem ipsum", is_active=True
        )

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"terms_of_service": ["This agreement is required."]}
        )

        # invalid agreement id
        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
                "terms_of_service": agreement.id + 1,
            },
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"terms_of_service": ["This agreement is required."]}
        )

        # valid agreement id
        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
                "terms_of_service": agreement.id,
            },
        )

        self.assertEqual(response.status_code, 200)

        user = User.objects.get(email="loremipsum@dolor.met")
        self.assertEqual(user.agreements, [agreement.id])
        self.assertEqual(user.useragreement_set.count(), 1)

    def test_registration_ignore_inactive_agreement(self):
        """api ignores inactive agreement"""
        Agreement.objects.create(
            type=Agreement.TYPE_TOS, text="Lorem ipsum", is_active=False
        )

        response = self.client.post(
            self.api_link,
            data={
                "username": "totallyNew",
                "email": "loremipsum@dolor.met",
                "password": "LoremP4ssword",
                "terms_of_service": "",
            },
        )

        self.assertEqual(response.status_code, 200)

        user = User.objects.get(email="loremipsum@dolor.met")
        self.assertEqual(user.agreements, [])
        self.assertEqual(user.useragreement_set.count(), 0)

    def test_registration_calls_validate_new_registration(self):
        """api uses validate_new_registration to validate registrations"""
        response = self.client.post(
            self.api_link,
            data={
                "username": "User",
                "email": "l.o.r.e.m.i.p.s.u.m@gmail.com",
                "password": "pas123",
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "email": ["This e-mail is not allowed."],
                "password": [
                    "This password is too short. It must contain at least 7 characters."
                ],
            },
        )

    @override_dynamic_settings(account_activation="none")
    def test_registration_creates_active_user(self):
        """api creates active and signed in user on POST"""
        response = self.client.post(
            self.api_link,
            data={
                "username": "User",
                "email": "user@example.com",
                "password": self.USER_PASSWORD,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {"activation": "active", "username": "User", "email": "user@example.com"},
        )

        User.objects.get_by_username("User")

        test_user = User.objects.get_by_email("user@example.com")
        self.assertEqual(Online.objects.filter(user=test_user).count(), 1)

        self.assertTrue(test_user.check_password(self.USER_PASSWORD))

        auth_json = self.client.get(reverse("misago:api:auth")).json()
        self.assertTrue(auth_json["is_authenticated"])
        self.assertEqual(auth_json["username"], "User")

        self.assertIn("Welcome", mail.outbox[0].subject)

        self.assertEqual(test_user.audittrail_set.count(), 1)

    @override_dynamic_settings(account_activation="user")
    def test_registration_creates_inactive_user(self):
        """api creates inactive user on POST"""
        response = self.client.post(
            self.api_link,
            data={
                "username": "User",
                "email": "user@example.com",
                "password": self.USER_PASSWORD,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {"activation": "user", "username": "User", "email": "user@example.com"},
        )

        auth_json = self.client.get(reverse("misago:api:auth")).json()
        self.assertFalse(auth_json["is_authenticated"])

        User.objects.get_by_username("User")
        User.objects.get_by_email("user@example.com")

        self.assertIn("Welcome", mail.outbox[0].subject)

    @override_dynamic_settings(account_activation="admin")
    def test_registration_creates_admin_activated_user(self):
        """api creates admin activated user on POST"""
        response = self.client.post(
            self.api_link,
            data={
                "username": "User",
                "email": "user@example.com",
                "password": self.USER_PASSWORD,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {"activation": "admin", "username": "User", "email": "user@example.com"},
        )

        auth_json = self.client.get(reverse("misago:api:auth")).json()
        self.assertFalse(auth_json["is_authenticated"])

        User.objects.get_by_username("User")
        User.objects.get_by_email("user@example.com")

        self.assertIn("Welcome", mail.outbox[0].subject)

    @override_dynamic_settings(account_activation="none")
    def test_registration_creates_user_with_whitespace_password(self):
        """api creates user with spaces around password"""
        password = " %s " % self.USER_PASSWORD
        response = self.client.post(
            self.api_link,
            data={
                "username": "User",
                "email": "user@example.com",
                "password": password,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {"activation": "active", "username": "User", "email": "user@example.com"},
        )

        User.objects.get_by_username("User")

        test_user = User.objects.get_by_email("user@example.com")
        self.assertEqual(Online.objects.filter(user=test_user).count(), 1)
        self.assertTrue(test_user.check_password(password))

        self.assertIn("Welcome", mail.outbox[0].subject)


@override_dynamic_settings(account_activation="none")
def test_new_registrations_validators_hook_is_used_by_registration_api(
    db, client, mocker
):
    def validator(request, cleaned_data, add_error):
        add_error("username", "ERROR FROM PLUGIN")

    mocker.patch(
        "misago.users.validators.hooks.new_registrations_validators", [validator]
    )

    response = client.post(
        reverse("misago:api:user-list"),
        {
            "username": "User",
            "email": "user@example.com",
            "password": "PASSW0RD123",
        },
    )

    assert response.status_code == 400
    assert response.json() == {"username": ["ERROR FROM PLUGIN"]}


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_registration_api_returns_403_if_oauth_is_enabled(db, client):
    response = client.post(
        reverse("misago:api:user-list"),
        {
            "username": "totallyNew",
            "email": "loremipsum@dolor.met",
            "password": "PASSW0RD123",
        },
    )
    assert response.status_code == 403



from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import pgettext_lazy

from ... import THREADS_ROOT_NAME
from ....acl.cache import clear_acl_cache
from ....admin.views import generic
from ....threads.threadtypes import trees_map
from ...models import Category, RoleCategoryACL
from ..forms import CategoryFormFactory, DeleteFormFactory


class CategoryAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:categories:index"
    model = Category
    templates_dir = "misago/admin/categories"
    message_404 = pgettext_lazy(
        "admin categories", "Requested category does not exist."
    )

    def get_target(self, kwargs):
        target = super().get_target(kwargs)

        threads_tree_id = trees_map.get_tree_id_for_root(THREADS_ROOT_NAME)

        target_is_special = bool(target.special_role)
        target_not_in_categories_tree = target.tree_id != threads_tree_id

        if target.pk and (target_is_special or target_not_in_categories_tree):
            raise Category.DoesNotExist()
        else:
            return target


class CategoriesList(CategoryAdmin, generic.ListView):
    def get_queryset(self):
        return Category.objects.all_categories()

    def process_context(self, request, context):
        context["items"] = [f for f in context["items"]]

        children_lists = {}

        for item in context["items"]:
            item.level_range = range(item.level - 1)
            item.first = False
            item.last = False
            children_lists.setdefault(item.parent_id, []).append(item)

        for level_items in children_lists.values():
            level_items[0].first = True
            level_items[-1].last = True

        return context


class CategoryFormMixin:
    def get_form_class(self, request, target):
        return CategoryFormFactory(target)

    def handle_form(self, form, request, target):
        if form.instance.pk:
            if form.instance.parent_id != form.cleaned_data["new_parent"].pk:
                form.instance.move_to(
                    form.cleaned_data["new_parent"], position="last-child"
                )
            form.instance.save()
            if form.instance.parent_id != form.cleaned_data["new_parent"].pk:
                Category.objects.clear_cache()
        else:
            form.instance.insert_at(
                form.cleaned_data["new_parent"], position="last-child", save=True
            )
            Category.objects.clear_cache()

        if form.cleaned_data.get("copy_permissions"):
            form.instance.category_role_set.all().delete()
            copy_from = form.cleaned_data["copy_permissions"]

            copied_acls = []
            for acl in copy_from.category_role_set.all():
                copied_acls.append(
                    RoleCategoryACL(
                        role_id=acl.role_id,
                        category=form.instance,
                        category_role_id=acl.category_role_id,
                    )
                )

            if copied_acls:
                RoleCategoryACL.objects.bulk_create(copied_acls)

        clear_acl_cache()
        messages.success(request, self.message_submit % {"name": target.name})


class NewCategory(CategoryFormMixin, CategoryAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy(
        "admin categories", 'New category "%(name)s" has been saved.'
    )


class EditCategory(CategoryFormMixin, CategoryAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy(
        "admin categories", 'Category "%(name)s" has been edited.'
    )


class DeleteCategory(CategoryAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy(
        "admin categories", 'Category "%(name)s" has been deleted.'
    )
    template_name = "delete.html"

    def get_form_class(self, request, target):
        return DeleteFormFactory(target)

    def handle_form(self, form, request, target):
        move_children_to = form.cleaned_data.get("move_children_to")
        move_threads_to = form.cleaned_data.get("move_threads_to")

        if move_children_to:
            for child in target.get_children():
                # refresh child and new parent
                move_children_to = Category.objects.get(pk=move_children_to.pk)
                child = Category.objects.get(pk=child.pk)

                child.move_to(move_children_to, "last-child")
                if move_threads_to and child.pk == move_threads_to.pk:
                    move_threads_to = child
        else:
            for child in target.get_descendants().order_by("-lft"):
                child.delete_content()
                child.delete()

        if move_threads_to:
            target.move_content(move_threads_to)
            move_threads_to.synchronize()
            move_threads_to.save()
        else:
            target.delete_content()

        # refresh instance
        instance = Category.objects.get(pk=form.instance.pk)
        instance.delete()

        messages.success(request, self.message_submit % {"name": target.name})
        return redirect(self.root_link)


class MoveDownCategory(CategoryAdmin, generic.ButtonView):
    def button_action(self, request, target):
        try:
            other_target = target.get_next_sibling()
        except Category.DoesNotExist:
            other_target = None

        if other_target:
            Category.objects.move_node(target, other_target, "right")
            Category.objects.clear_cache()

            message = pgettext_lazy(
                "admin categories",
                'Category "%(name)s" has been moved below "%(other)s".',
            )
            targets_names = {"name": target.name, "other": other_target.name}
            messages.success(request, message % targets_names)


class MoveUpCategory(CategoryAdmin, generic.ButtonView):
    def button_action(self, request, target):
        try:
            other_target = target.get_previous_sibling()
        except Category.DoesNotExist:
            other_target = None

        if other_target:
            Category.objects.move_node(target, other_target, "left")
            Category.objects.clear_cache()

            message = pgettext_lazy(
                "admin categories",
                'Category "%(name)s" has been moved above "%(other)s".',
            )
            targets_names = {"name": target.name, "other": other_target.name}
            messages.success(request, message % targets_names)



from django.core.exceptions import ValidationError
from django.urls import reverse

from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase


class ValidatePostTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.api_link = reverse("misago:api:thread-list")

    def test_title_validation(self):
        """validate_post tests title"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Check our l33t CaSiNo!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"non_field_errors": ["Don't discuss gambling!"]}
        )

        # clean title passes validation
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Check our l33t place!",
                "post": "Lorem ipsum dolor met!",
            },
        )
        self.assertEqual(response.status_code, 200)

    def test_post_validation(self):
        """validate_post tests post content"""
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Lorem ipsum dolor met!",
                "post": "Check our l33t CaSiNo!",
            },
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"non_field_errors": ["Don't discuss gambling!"]}
        )

        # clean post passes validation
        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.pk,
                "title": "Lorem ipsum dolor met!",
                "post": "Check our l33t place!",
            },
        )
        self.assertEqual(response.status_code, 200)

    def test_empty_input(self):
        """validate_post handles empty input"""
        response = self.client.post(self.api_link, data={"category": self.category.pk})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "title": ["You have to enter a thread title."],
                "post": ["You have to enter a message."],
            },
        )

        response = self.client.post(
            self.api_link, data={"category": self.category.pk, "title": "", "post": ""}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "title": ["This field may not be blank."],
                "post": ["This field may not be blank."],
            },
        )


def test_post_validators_hook_is_used_by_posting_api(db, user_client, mocker, thread):
    def validator(context, data):
        raise ValidationError("ERROR FROM PLUGIN")

    mocker.patch("misago.threads.validators.hooks.post_validators", [validator])

    response = user_client.post(
        f"/api/threads/{thread.id}/posts/", {"post": "Lorem ipsum dolor met"}
    )

    assert response.status_code == 400
    assert response.json() == {"non_field_errors": ["ERROR FROM PLUGIN"]}



from django import template
from django.urls import NoReverseMatch, reverse

register = template.Library()


@register.simple_tag(takes_context=True)
def absoluteurl(context, url_or_name, *args, **kwargs):
    address = context["settings"].forum_address
    if not address:
        return None

    absolute_url_prefix = address.rstrip("/")

    try:
        url_or_name = reverse(url_or_name, args=args, kwargs=kwargs)
    except NoReverseMatch:
        # don't use URLValidator because its too explicit
        if not url_or_name.startswith("/"):
            return url_or_name

    return "%s%s" % (absolute_url_prefix, url_or_name)



from .categories_map import get_categories_map


def preload_categories_json(request):
    if not request.include_frontend_context:
        return {}

    request.frontend_context.update({"categoriesMap": get_categories_map(request)})

    return {}



from functools import wraps

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect
from django.utils.translation import pgettext

from ..core.exceptions import Banned
from .bans import get_request_ip_ban
from .models import Ban


def deny_authenticated(f):
    @wraps(f)
    def decorator(request, *args, **kwargs):
        if request.user.is_authenticated:
            raise PermissionDenied(
                pgettext(
                    "block authenticated decorator",
                    "This page is not available to signed in users.",
                )
            )
        else:
            return f(request, *args, **kwargs)

    return decorator


def deny_guests(f):
    @wraps(f)
    def decorator(request, *args, **kwargs):
        if request.user.is_anonymous:
            if request.GET.get("ref") == "login":
                return redirect(settings.LOGIN_REDIRECT_URL)
            raise PermissionDenied(
                pgettext(
                    "block guests decorator", "You have to sign in to access this page."
                )
            )
        else:
            return f(request, *args, **kwargs)

    return decorator


def deny_banned_ips(f):
    @wraps(f)
    def decorator(request, *args, **kwargs):
        ban = get_request_ip_ban(request)
        if ban:
            hydrated_ban = Ban(
                check_type=Ban.IP,
                user_message=ban["message"],
                expires_on=ban["expires_on"],
            )
            raise Banned(hydrated_ban)
        else:
            return f(request, *args, **kwargs)

    return decorator



import copy
import logging

from django.core.exceptions import ValidationError
from django.urls import reverse
from django.utils.module_loading import import_string
from django.utils.translation import gettext

from ...conf import settings

from .basefields import *
from .serializers import serialize_profilefields_data


logger = logging.getLogger("misago.users.ProfileFields")


class ProfileFields:
    def __init__(self, fields_groups):
        self.is_loaded = False

        self.fields_groups = fields_groups
        self.fields_dict = {}

    def load(self):
        self.fields_dict = {}

        fieldnames = {}

        for group in self.fields_groups:
            for field_path in group["fields"]:
                field = import_string(field_path)
                field._field_path = field_path

                if field_path in self.fields_dict:
                    raise ValueError(
                        "%s profile field has been specified twice" % field._field_path
                    )

                if not getattr(field, "fieldname", None):
                    raise ValueError(
                        "%s profile field has to specify fieldname attribute"
                        % field._field_path
                    )

                if field.fieldname in fieldnames:
                    raise ValueError(
                        (
                            '%s profile field defines fieldname "%s" '
                            "that is already in use by the %s"
                        )
                        % (
                            field._field_path,
                            field.fieldname,
                            fieldnames[field.fieldname],
                        )
                    )

                fieldnames[field.fieldname] = field_path
                self.fields_dict[field_path] = field()

        self.is_loaded = True

    def get_fields(self):
        if not self.is_loaded:
            self.load()
        return self.fields_dict.values()

    def get_fields_groups(self):
        if not self.is_loaded:
            self.load()

        groups = []
        for group in self.fields_groups:
            group_dict = {"name": gettext(group["name"]), "fields": []}

            for field_path in group["fields"]:
                field = self.fields_dict[field_path]
                group_dict["fields"].append(field)

            if group_dict["fields"]:
                groups.append(group_dict)
        return groups

    def add_fields_to_form(self, request, user, form):
        if not self.is_loaded:
            self.load()

        form._profile_fields = []

        for field in self.get_fields():
            if not field.is_editable(request, user):
                continue

            form._profile_fields.append(field.fieldname)
            form.fields[field.fieldname] = field.get_form_field(request, user)

    def add_fields_to_admin_form(self, request, user, form):
        self.add_fields_to_form(request, user, form)

        form._profile_fields_groups = []
        for group in self.fields_groups:
            group_dict = {"name": gettext(group["name"]), "fields": []}

            for field_path in group["fields"]:
                field = self.fields_dict[field_path]
                if field.fieldname in form._profile_fields:
                    group_dict["fields"].append(field.fieldname)

            if group_dict["fields"]:
                form._profile_fields_groups.append(group_dict)

    def clean_form(self, request, user, form, cleaned_data):
        for field in self.get_fields():
            if field.fieldname not in cleaned_data:
                continue

            try:
                cleaned_data[field.fieldname] = field.clean(
                    request, user, cleaned_data[field.fieldname]
                )
            except ValidationError as e:
                form.add_error(field.fieldname, e)

        return cleaned_data

    def update_user_profile_fields(self, request, user, form):
        old_fields = copy.copy(user.profile_fields or {})

        new_fields = {}
        for fieldname in form._profile_fields:
            if fieldname in form.cleaned_data:
                new_fields[fieldname] = form.cleaned_data[fieldname]
        user.profile_fields = new_fields

        old_fields_reduced = {k: v for k, v in old_fields.items() if v}
        new_fields_reduced = {k: v for k, v in new_fields.items() if v}

        if old_fields_reduced != new_fields_reduced:
            self.log_profile_fields_update(request, user)

    def log_profile_fields_update(self, request, user):
        if request.user == user:
            log_message = "%s edited own profile fields" % user.username
        else:
            log_message = "%s edited %s's (#%s) profile fields" % (
                request.user,
                user.username,
                user.pk,
            )

        logger.info(
            log_message,
            extra={
                "absolute_url": request.build_absolute_uri(
                    reverse(
                        "misago:user-details", kwargs={"slug": user.slug, "pk": user.pk}
                    )
                )
            },
        )

    def search_users(self, criteria, queryset):
        if not self.is_loaded:
            self.load()

        q_obj = None
        for field in self.fields_dict.values():
            q = field.search_users(criteria)
            if q:
                if q_obj:
                    q_obj = q_obj | q
                else:
                    q_obj = q
        if q_obj:
            return queryset.filter(q_obj)

        return queryset


profilefields = ProfileFields(settings.MISAGO_PROFILE_FIELDS)



from datetime import timedelta
from io import StringIO

import pytest
from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.utils import timezone

from ...categories.models import Category
from ...threads.models import Post, Thread
from ...users.models import Rank
from ..management.commands import createfakehistory
from ..threads import get_fake_thread
from ..users import get_fake_admin_activated_user, get_fake_inactive_user, get_fake_user

User = get_user_model()


@pytest.fixture
def command(db):
    return createfakehistory.Command(stdout=StringIO())


@pytest.fixture
def date():
    return timezone.now()


def test_management_command_has_no_errors(db):
    call_command(createfakehistory.Command(), max_actions=3, stdout=StringIO())


def test_management_command_creates_fake_user(fake, command, date):
    ranks = list(Rank.objects.all())
    command.create_fake_user(fake, date, ranks)
    assert User.objects.exists()


def test_fake_user_join_date_is_overridden_by_command(fake, command, date):
    ranks = list(Rank.objects.all())
    command.create_fake_user(fake, date, ranks)
    user = User.objects.order_by("-pk").last()
    assert user.joined_on == date


def test_fake_user_rank_is_one_from_the_choices(fake, command, date):
    ranks = list(Rank.objects.all())
    command.create_fake_user(fake, date, ranks)
    user = User.objects.order_by("-pk").last()
    assert user.rank in ranks


def test_none_is_returned_for_random_user_if_no_users_exist(command, date):
    user = command.get_random_user(date)
    assert user is None


def test_users_created_after_given_date_are_excluded_from_random_user_pick(
    command, date, other_user
):
    other_user.joined_on = timezone.now()
    other_user.save()

    user = command.get_random_user(date)
    assert user is None


def test_inactive_users_are_excluded_from_random_user_pick(fake, command):
    get_fake_admin_activated_user(fake)
    get_fake_inactive_user(fake)

    user = command.get_random_user(timezone.now())
    assert user is None


def test_random_user_pick_returns_random_user(fake, command):
    valid_choices = [get_fake_user(fake) for _ in range(5)]
    user = command.get_random_user(timezone.now())
    assert user in valid_choices


def test_management_command_creates_fake_thread(fake, command, date):
    categories = list(Category.objects.all_categories())
    command.create_fake_thread(fake, date, categories)
    assert Thread.objects.exists()


def test_fake_thread_start_date_is_overridden_by_command(fake, command, date):
    categories = list(Category.objects.all_categories())
    command.create_fake_thread(fake, date, categories)
    thread = Thread.objects.last()
    assert thread.started_on == date


def test_fake_thread_was_created_in_one_of_valid_categories(fake, command, date):
    categories = list(Category.objects.all_categories())
    command.create_fake_thread(fake, date, categories)
    thread = Thread.objects.last()
    assert thread.category in categories


def test_none_is_returned_for_random_thread_if_no_threads_exist(command, date):
    thread = command.get_random_thread(date)
    assert thread is None


def test_threads_created_after_given_date_are_excluded_from_random_thread_pick(
    fake, command, date, default_category
):
    get_fake_thread(fake, default_category)
    thread = command.get_random_thread(date)
    assert thread is None


def test_random_thread_pick_returns_random_thread(fake, command, default_category):
    valid_choices = [get_fake_thread(fake, default_category) for _ in range(5)]
    thread = command.get_random_thread(timezone.now())
    assert thread in valid_choices


def test_management_command_creates_fake_post(fake, command, default_category):
    thread = get_fake_thread(fake, default_category)
    command.create_fake_post(fake, timezone.now())
    assert thread.post_set.count() == 2


def test_fake_post_creation_date_is_overridden_by_command(
    fake, command, date, default_category
):
    thread = get_fake_thread(fake, default_category)
    thread.started_on -= timedelta(days=1)
    thread.save()

    command.create_fake_post(fake, date)
    post = thread.post_set.last()
    assert post.posted_on == date


def test_fake_post_is_not_created_if_no_threads_exist(fake, command, date):
    command.create_fake_post(fake, date)
    assert not Post.objects.exists()


def test_management_command_synchronizes_threads(fake, command, date, default_category):
    command.create_fake_thread(fake, date, [default_category])
    command.synchronize_threads()


def test_management_command_synchronizes_categories(
    fake, command, date, default_category
):
    command.create_fake_thread(fake, date, [default_category])
    command.synchronize_threads()
    command.synchronize_categories()



from django.db.models import F
from django.http import HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404, redirect, render

from .models import Notification, WatchedThread
from .permissions import allow_use_notifications
from .registry import registry


def notifications(request: HttpRequest) -> HttpResponse:
    allow_use_notifications(request.user)
    return render(request, "misago/notifications.html")


def notification(request: HttpRequest, notification_id: int) -> HttpResponse:
    allow_use_notifications(request.user)

    user = request.user
    notification = get_object_or_404(
        Notification.objects.select_related(
            "actor",
            "category",
            "thread",
            "post",
        ),
        user=user,
        id=notification_id,
    )

    # Populate user relation cache
    notification.user = request.user

    if notification.category_id:
        # Populate relations caches on thread and post models
        if notification.thread_id:
            notification.thread.category = notification.category
        if notification.post_id:
            notification.post.thread = notification.thread
            notification.post.category = notification.category

    if not notification.is_read:
        notification.is_read = True
        notification.save(update_fields=["is_read"])

        if user.unread_notifications:
            user.unread_notifications = F("unread_notifications") - 1
            user.save(update_fields=["unread_notifications"])

    return redirect(registry.get_redirect_url(request, notification))


def disable_email_notifications(
    request: HttpRequest, watched_thread_id: int, secret: str
) -> HttpResponse:
    watched_thread = get_object_or_404(
        WatchedThread.objects.select_related("user", "category", "thread"),
        id=watched_thread_id,
        secret=secret,
    )

    if watched_thread.send_emails:
        watched_thread.send_emails = False
        watched_thread.save(update_fields=["send_emails"])

    watched_thread.thread.category = watched_thread.category

    return render(
        request,
        "misago/notifications_disabled.html",
        {
            "watcher": watched_thread.user,
            "thread": watched_thread.thread,
        },
    )



from .. import discover
from ..manifest import MisagoPlugin
from ..metadata import (
    clean_plugin_color,
    clean_plugin_icon,
    clean_plugin_str,
    clean_plugin_url,
    create_plugin_metadata,
)


def test_plugin_metadata_is_created_without_manifest():
    metadata = create_plugin_metadata("plugin_package", discover, None, None)

    assert metadata.package == "plugin_package"
    assert metadata.dirname == "misago"
    assert metadata.has_manifest is False
    assert metadata.manifest_error is None
    assert metadata.name is None
    assert metadata.description is None
    assert metadata.license is None
    assert metadata.icon is None
    assert metadata.color is None
    assert metadata.version is None
    assert metadata.author is None
    assert metadata.homepage is None
    assert metadata.sponsor is None
    assert metadata.help is None
    assert metadata.bugs is None
    assert metadata.repo is None


def test_plugin_metadata_is_created_with_manifest_error():
    metadata = create_plugin_metadata(
        "plugin_package", discover, None, "Not found error."
    )

    assert metadata.package == "plugin_package"
    assert metadata.dirname == "misago"
    assert metadata.has_manifest is False
    assert metadata.manifest_error == "Not found error."
    assert metadata.name is None
    assert metadata.description is None
    assert metadata.license is None
    assert metadata.icon is None
    assert metadata.color is None
    assert metadata.version is None
    assert metadata.author is None
    assert metadata.homepage is None
    assert metadata.sponsor is None
    assert metadata.help is None
    assert metadata.bugs is None
    assert metadata.repo is None


def test_plugin_metadata_is_created_from_empty_manifest():
    manifest = MisagoPlugin()
    metadata = create_plugin_metadata(
        "plugin_package",
        discover,
        manifest,
        None,
    )

    assert metadata.package == "plugin_package"
    assert metadata.dirname == "misago"
    assert metadata.has_manifest is True
    assert metadata.manifest_error is None
    assert metadata.name is None
    assert metadata.description is None
    assert metadata.license is None
    assert metadata.icon is None
    assert metadata.color is None
    assert metadata.version is None
    assert metadata.author is None
    assert metadata.homepage is None
    assert metadata.sponsor is None
    assert metadata.help is None
    assert metadata.bugs is None
    assert metadata.repo is None


def test_plugin_metadata_is_created_from_valid_full_manifest():
    manifest = MisagoPlugin(
        name="Test plugin",
        description="Test plugin's manifest",
        license="GNU GPL v2",
        icon="fa fa-puzzle",
        color="#FF00F0",
        version="1.0",
        author="Lorem Ipsum",
        homepage="https://misago-project.org",
        sponsor="https://sponsor.misago-project.org",
        help="https://help.misago-project.org",
        bugs="https://bugs.misago-project.org",
        repo="https://repo.misago-project.org",
    )

    metadata = create_plugin_metadata(
        "plugin_package",
        discover,
        manifest,
        None,
    )

    assert metadata.package == "plugin_package"
    assert metadata.dirname == "misago"
    assert metadata.has_manifest is True
    assert metadata.manifest_error is None
    assert metadata.name == "Test plugin"
    assert metadata.description == "Test plugin's manifest"
    assert metadata.license == "GNU GPL v2"
    assert metadata.icon == "fa fa-puzzle"
    assert metadata.color == "#FF00F0"
    assert metadata.version == "1.0"
    assert metadata.author == "Lorem Ipsum"
    assert metadata.homepage.netloc == "misago-project.org"
    assert metadata.homepage.url == "https://misago-project.org"
    assert metadata.sponsor.netloc == "sponsor.misago-project.org"
    assert metadata.sponsor.url == "https://sponsor.misago-project.org"
    assert metadata.help.netloc == "help.misago-project.org"
    assert metadata.help.url == "https://help.misago-project.org"
    assert metadata.bugs.netloc == "bugs.misago-project.org"
    assert metadata.bugs.url == "https://bugs.misago-project.org"
    assert metadata.repo.netloc == "repo.misago-project.org"
    assert metadata.repo.url == "https://repo.misago-project.org"


def test_plugin_color_returns_none_if_manifest_value_is_none():
    assert clean_plugin_color(None) is None


def test_plugin_color_returns_none_if_manifest_value_is_not_a_str():
    assert clean_plugin_color([1, 2, 3]) is None
    assert clean_plugin_color({}) is None
    assert clean_plugin_color(False) is None
    assert clean_plugin_color(42) is None


def test_plugin_color_returns_none_if_manifest_value_is_invalid_str():
    assert clean_plugin_color("") is None
    assert clean_plugin_color("red") is None
    assert clean_plugin_color("#8823412") is None
    assert clean_plugin_color("#zzzzzz") is None
    assert clean_plugin_color("eke eke nii") is None


def test_plugin_color_returns_valid_manifest_color():
    assert clean_plugin_color("#ffffff") == "#FFFFFF"
    assert clean_plugin_color("#000000") == "#000000"


def test_plugin_icon_returns_none_if_manifest_value_is_none():
    assert clean_plugin_icon(None) is None


def test_plugin_icon_returns_none_if_manifest_value_is_not_a_str():
    assert clean_plugin_icon([1, 2, 3]) is None
    assert clean_plugin_icon({}) is None
    assert clean_plugin_icon(False) is None
    assert clean_plugin_icon(42) is None


def test_plugin_icon_returns_none_if_manifest_value_is_invalid_str():
    assert clean_plugin_icon("") is None
    assert clean_plugin_icon("red") is None
    assert clean_plugin_icon("bell clock") is None
    assert clean_plugin_icon("fas other") is None


def test_plugin_icon_returns_valid_manifest_icon():
    assert clean_plugin_icon("fa fa-test") == "fa fa-test"
    assert clean_plugin_icon("fas fa-other-test") == "fas fa-other-test"
    assert clean_plugin_icon("far fa-extra-test") == "far fa-extra-test"


def test_plugin_str_returns_none_if_manifest_value_is_none():
    assert clean_plugin_str(None, 10) is None


def test_plugin_str_returns_none_if_manifest_value_is_not_a_str():
    assert clean_plugin_str([1, 2, 3], 10) is None
    assert clean_plugin_str({}, 10) is None
    assert clean_plugin_str(False, 10) is None
    assert clean_plugin_str(42, 10) is None


def test_plugin_str_returns_none_if_manifest_value_is_empty_str():
    assert clean_plugin_str("", 10) is None
    assert clean_plugin_str("  ", 10) is None


def test_plugin_str_returns_string_cut_to_length():
    assert clean_plugin_str("Hello world", 15) == "Hello world"
    assert clean_plugin_str("Hello world", 5) == "Hello"


def test_plugin_str_returns_string_with_whitespace_stripped():
    assert clean_plugin_str("    Hello world   ", 15) == "Hello world"
    assert clean_plugin_str("  Hello world  ", 5) == "Hello"


def test_plugin_url_returns_none_if_manifest_value_is_none():
    assert clean_plugin_url(None) is None


def test_plugin_url_returns_none_if_manifest_value_is_not_a_str():
    assert clean_plugin_url([1, 2, 3]) is None
    assert clean_plugin_url({}) is None
    assert clean_plugin_url(False) is None
    assert clean_plugin_url(42) is None


def test_plugin_url_returns_none_if_manifest_value_is_invalid_str():
    assert clean_plugin_url("") is None
    assert clean_plugin_url("red") is None
    assert clean_plugin_url("misago-project.org") is None
    assert clean_plugin_url("www.misago-project.org") is None
    assert clean_plugin_url('javascript:alert("test")') is None
    assert clean_plugin_url("mailto:dot@exc.com") is None
    assert clean_plugin_url("ftp://oke.com/guru") is None


def test_plugin_url_returns_url_if_manifest_value_is_supported_url():
    http_url = clean_plugin_url("http://example.com")
    assert http_url.netloc == "example.com"
    assert http_url.url == "http://example.com"

    https_url = clean_plugin_url("https://example.com")
    assert https_url.netloc == "example.com"
    assert https_url.url == "https://example.com"

    comma_url = clean_plugin_url("https://example-site.com")
    assert comma_url.netloc == "example-site.com"
    assert comma_url.url == "https://example-site.com"

    digits_url = clean_plugin_url("https://example-site365.com")
    assert digits_url.netloc == "example-site365.com"
    assert digits_url.url == "https://example-site365.com"

    https_www_url = clean_plugin_url("https://www.example.com")
    assert https_www_url.netloc == "www.example.com"
    assert https_www_url.url == "https://www.example.com"

    https_www_subdomains_url = clean_plugin_url("https://www.dev.stage.example.com")
    assert https_www_subdomains_url.netloc == "www.dev.stage.example.com"
    assert https_www_subdomains_url.url == "https://www.dev.stage.example.com"

    http_php_url = clean_plugin_url("http://example.com/some/where.php")
    assert http_php_url.netloc == "example.com"
    assert http_php_url.url == "http://example.com/some/where.php"

    http_path_url = clean_plugin_url("http://example.com/some-path/ok/123")
    assert http_path_url.netloc == "example.com"
    assert http_path_url.url == "http://example.com/some-path/ok/123"

    http_fragment_url = clean_plugin_url("http://example.com/some-path/#ok-123")
    assert http_fragment_url.netloc == "example.com"
    assert http_fragment_url.url == "http://example.com/some-path/#ok-123"

    http_query_url = clean_plugin_url("http://example.com/arg/?ok=yes&alt=no")
    assert http_query_url.netloc == "example.com"
    assert http_query_url.url == "http://example.com/arg/?ok=yes&alt=no"



from datetime import timedelta

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.utils import timezone

from ....conf.shortcuts import get_dynamic_settings
from ....core.pgutils import chunk_queryset
from ...deletesrecord import record_user_deleted_by_system

User = get_user_model()


class Command(BaseCommand):
    help = "Deletes inactive user accounts older than set time."

    def handle(self, *args, **options):
        settings = get_dynamic_settings()
        if not settings.new_inactive_accounts_delete:
            self.stdout.write(
                "Automatic deletion of inactive user accounts is currently disabled."
            )
            return

        users_deleted = 0

        joined_on_cutoff = timezone.now() - timedelta(
            days=settings.new_inactive_accounts_delete
        )

        queryset = User.objects.filter(
            requires_activation__gt=User.ACTIVATION_NONE, joined_on__lt=joined_on_cutoff
        )

        for user in chunk_queryset(queryset):
            user.delete(anonymous_username=settings.anonymous_username)
            record_user_deleted_by_system()
            users_deleted += 1

        self.stdout.write("Deleted inactive user accounts: %s" % users_deleted)



from ...test import assert_contains


def test_active_theme_styles_are_included_in_page_html(client, active_theme):
    css = active_theme.css.create(name="test", url="https://cdn.example.com/style.css")
    response = client.get("/")
    assert_contains(response, 'link href="%s" rel="stylesheet"' % css.url)



from django.utils import timezone
from rest_framework.request import Request

from ..models import Online


def mute_tracker(request):
    request._misago_online_tracker = None


def start_tracking(request, user):
    online_tracker = Online.objects.create(user=user)

    request.user.online_tracker = online_tracker
    request._misago_online_tracker = online_tracker


def update_tracker(request, tracker):
    tracker.last_click = timezone.now()

    tracker.save(update_fields=["last_click"])


def stop_tracking(request, tracker):
    user = tracker.user
    user.last_login = tracker.last_click
    user.save(update_fields=["last_login"])

    tracker.delete()


def clear_tracking(request):
    if isinstance(request, Request):
        request = request._request  # Fugly unwrap restframework's request
    request._misago_online_tracker = None



from django.urls import reverse

from ....acl.test import patch_user_acl
from ....notifications.models import WatchedThread
from ....notifications.threads import ThreadNotifications


def test_thread_watch_api_doesnt_create_watched_thread_for_disabled_notifications(
    user, thread, user_client
):
    response = user_client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
        json={"notifications": ThreadNotifications.NONE.value},
    )

    assert response.status_code == 200
    assert response.json() == {"notifications": ThreadNotifications.NONE.value}

    assert not WatchedThread.objects.exists()


def test_thread_watch_api_creates_watched_thread_without_email_notifications(
    user, thread, user_client
):
    response = user_client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
        json={"notifications": ThreadNotifications.SITE_ONLY.value},
    )

    assert response.status_code == 200
    assert response.json() == {"notifications": ThreadNotifications.SITE_ONLY.value}

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert watched_thread.category_id == thread.category_id
    assert not watched_thread.send_emails


def test_thread_watch_api_creates_watched_thread_with_email_notifications(
    user, thread, user_client
):
    response = user_client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
        json={"notifications": ThreadNotifications.SITE_AND_EMAIL.value},
    )

    assert response.status_code == 200
    assert response.json() == {
        "notifications": ThreadNotifications.SITE_AND_EMAIL.value
    }

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert watched_thread.category_id == thread.category_id
    assert watched_thread.send_emails


def test_thread_watch_api_enables_watched_thread_email_notifications(
    user, thread, user_client, django_assert_num_queries
):
    existing_watched_thread = WatchedThread.objects.create(
        user=user,
        category_id=thread.category_id,
        thread=thread,
        send_emails=False,
    )

    with django_assert_num_queries(25):
        response = user_client.post(
            reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
            json={"notifications": ThreadNotifications.SITE_AND_EMAIL.value},
        )

    assert response.status_code == 200
    assert response.json() == {
        "notifications": ThreadNotifications.SITE_AND_EMAIL.value
    }

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert watched_thread.id == existing_watched_thread.id
    assert watched_thread.send_emails


def test_thread_watch_api_disables_watched_thread_email_notifications(
    user, thread, user_client, django_assert_num_queries
):
    existing_watched_thread = WatchedThread.objects.create(
        user=user,
        category_id=thread.category_id,
        thread=thread,
        send_emails=True,
    )

    with django_assert_num_queries(25):
        response = user_client.post(
            reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
            json={"notifications": ThreadNotifications.SITE_ONLY.value},
        )

    assert response.status_code == 200
    assert response.json() == {"notifications": ThreadNotifications.SITE_ONLY.value}

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert watched_thread.id == existing_watched_thread.id
    assert not watched_thread.send_emails


def test_thread_watch_api_skips_update_if_notifications_are_not_changed(
    user, thread, user_client, django_assert_num_queries
):
    existing_watched_thread = WatchedThread.objects.create(
        user=user,
        category_id=thread.category_id,
        thread=thread,
        send_emails=True,
    )

    with django_assert_num_queries(24):
        response = user_client.post(
            reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
            json={"notifications": ThreadNotifications.SITE_AND_EMAIL.value},
        )

    assert response.status_code == 200
    assert response.json() == {
        "notifications": ThreadNotifications.SITE_AND_EMAIL.value
    }

    watched_thread = WatchedThread.objects.get(user=user, thread=thread)
    assert watched_thread.id == existing_watched_thread.id
    assert watched_thread.send_emails


def test_thread_watch_api_deletes_watched_thread_if_notifications_are_disabled(
    user, thread, user_client
):
    WatchedThread.objects.create(
        user=user,
        category_id=thread.category_id,
        thread=thread,
        send_emails=True,
    )

    response = user_client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
        json={"notifications": ThreadNotifications.NONE.value},
    )

    assert response.status_code == 200
    assert response.json() == {"notifications": ThreadNotifications.NONE.value}

    assert not WatchedThread.objects.exists()


def test_thread_watch_api_returns_404_error_if_thread_doesnt_exist(user, user_client):
    response = user_client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": 404}),
        json={"notifications": ThreadNotifications.SITE_AND_EMAIL.value},
    )

    assert response.status_code == 404
    assert WatchedThread.objects.count() == 0


def test_thread_watch_api_returns_404_error_if_thread_is_not_accessible(
    user, other_user_hidden_thread, user_client
):
    response = user_client.post(
        reverse(
            "misago:apiv2:thread-watch",
            kwargs={"thread_id": other_user_hidden_thread.id},
        ),
        json={"notifications": ThreadNotifications.SITE_AND_EMAIL.value},
    )

    assert response.status_code == 404
    assert WatchedThread.objects.count() == 0


def test_thread_watch_api_returns_400_error_if_notifications_are_not_provided(
    user, thread, user_client
):
    response = user_client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}), json={}
    )

    assert response.status_code == 400
    assert response.json() == {"notifications": ["This field is required."]}

    assert WatchedThread.objects.count() == 0


def test_thread_watch_api_returns_400_error_if_notifications_are_invalid(
    user, thread, user_client
):
    response = user_client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
        json={"notifications": 42},
    )

    assert response.status_code == 400
    assert response.json() == {"notifications": ['"42" is not a valid choice.']}

    assert WatchedThread.objects.count() == 0


def test_thread_watch_api_returns_403_error_if_client_is_no_authenticated(
    thread, client
):
    response = client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": thread.id}),
        json={"notifications": ThreadNotifications.SITE_AND_EMAIL.value},
    )

    assert response.status_code == 403
    assert WatchedThread.objects.count() == 0


def test_thread_watch_api_returns_404_error_if_thread_is_private(
    user, private_thread, user_client
):
    response = user_client.post(
        reverse("misago:apiv2:thread-watch", kwargs={"thread_id": private_thread.id}),
        json={"notifications": ThreadNotifications.SITE_AND_EMAIL.value},
    )

    assert response.status_code == 404
    assert WatchedThread.objects.count() == 0


def test_private_thread_watch_api_creates_watched_thread_with_notifications(
    user, private_thread, user_client
):
    private_thread.threadparticipant_set.create(user=user)

    response = user_client.post(
        reverse(
            "misago:apiv2:private-thread-watch", kwargs={"thread_id": private_thread.id}
        ),
        json={"notifications": ThreadNotifications.SITE_ONLY.value},
    )

    assert response.status_code == 200
    assert response.json() == {"notifications": ThreadNotifications.SITE_ONLY.value}

    watched_thread = WatchedThread.objects.get(user=user, thread=private_thread)
    assert watched_thread.category_id == private_thread.category_id
    assert not watched_thread.send_emails


def test_private_thread_watch_api_returns_404_error_if_user_has_no_access_to_thread(
    user, private_thread, user_client
):
    response = user_client.post(
        reverse(
            "misago:apiv2:private-thread-watch", kwargs={"thread_id": private_thread.id}
        ),
        json={"notifications": ThreadNotifications.SITE_ONLY.value},
    )

    assert response.status_code == 404
    assert WatchedThread.objects.count() == 0


def test_private_thread_watch_api_returns_404_error_if_thread_is_not_private_thread(
    user, thread, user_client
):
    response = user_client.post(
        reverse("misago:apiv2:private-thread-watch", kwargs={"thread_id": thread.id}),
        json={"notifications": ThreadNotifications.SITE_ONLY.value},
    )

    assert response.status_code == 404
    assert WatchedThread.objects.count() == 0


def test_private_thread_watch_api_returns_403_error_if_client_is_no_authenticated(
    user, private_thread, client
):
    private_thread.threadparticipant_set.create(user=user)

    response = client.post(
        reverse(
            "misago:apiv2:private-thread-watch", kwargs={"thread_id": private_thread.id}
        ),
        json={"notifications": ThreadNotifications.SITE_ONLY.value},
    )

    assert response.status_code == 403
    assert WatchedThread.objects.count() == 0


@patch_user_acl({"can_use_private_threads": False})
def test_private_thread_watch_api_returns_403_if_user_cant_use_private_threads(
    user, private_thread, user_client
):
    private_thread.threadparticipant_set.create(user=user)

    response = user_client.post(
        reverse(
            "misago:apiv2:private-thread-watch", kwargs={"thread_id": private_thread.id}
        ),
        json={"notifications": ThreadNotifications.SITE_ONLY.value},
    )

    assert response.status_code == 403
    assert WatchedThread.objects.count() == 0



import pytest
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ....test import assert_contains, assert_has_error_message
from ... import THEME_CACHE
from ...models import Theme


@pytest.fixture
def create_link():
    return reverse("misago:admin:themes:new")


@pytest.fixture
def edit_link(theme):
    return reverse("misago:admin:themes:edit", kwargs={"pk": theme.pk})


def test_theme_creation_form_is_displayed(admin_client, create_link):
    response = admin_client.get(create_link)
    assert response.status_code == 200
    assert_contains(response, "New theme")


def test_theme_creation_form_reads_parent_from_url_and_preselects_it_in_parent_select(
    admin_client, create_link, theme
):
    response = admin_client.get("%s?parent=%s" % (create_link, theme.pk))
    assert_contains(response, '<option value="%s" selected>' % theme.pk)


def test_theme_creation_form_reads_parent_from_url_and_discards_invalid_value(
    admin_client, create_link, theme
):
    response = admin_client.get("%s?parent=%s" % (create_link, theme.pk + 1))
    assert response.status_code == 200


def test_theme_can_be_created(admin_client, create_link):
    admin_client.post(create_link, {"name": "New Theme"})
    Theme.objects.get(name="New Theme")


def test_child_theme_for_custom_theme_can_be_created(admin_client, create_link, theme):
    admin_client.post(create_link, {"name": "New Theme", "parent": theme.pk})
    child_theme = Theme.objects.get(name="New Theme")
    assert child_theme.parent == theme


def test_child_theme_for_default_theme_can_be_created(
    admin_client, create_link, default_theme
):
    admin_client.post(create_link, {"name": "New Theme", "parent": default_theme.pk})
    child_theme = Theme.objects.get(name="New Theme")
    assert child_theme.parent == default_theme


def test_creating_child_theme_updates_themes_tree(admin_client, create_link, theme):
    admin_client.post(create_link, {"name": "New Theme", "parent": theme.pk})

    child_theme = Theme.objects.get(name="New Theme")
    assert child_theme.tree_id == theme.tree_id
    assert child_theme.lft == 2
    assert child_theme.rght == 3

    theme.refresh_from_db()
    assert theme.lft == 1
    assert theme.rght == 4


def test_theme_creation_fails_if_name_is_not_given(admin_client, create_link):
    admin_client.post(create_link, {"name": ""})
    assert Theme.objects.count() == 1


def test_theme_creation_fails_if_parent_theme_doesnt_exist(
    admin_client, create_link, nonexisting_theme
):
    admin_client.post(
        create_link, {"name": "New Theme", "parent": nonexisting_theme.pk}
    )
    assert Theme.objects.count() == 1


def test_theme_edition_form_is_displayed(admin_client, edit_link, theme):
    response = admin_client.get(edit_link)
    assert response.status_code == 200
    assert_contains(response, theme.name)


def test_theme_name_can_be_edited(admin_client, edit_link, theme):
    admin_client.post(edit_link, {"name": "Edited Theme"})
    theme.refresh_from_db()
    assert theme.name == "Edited Theme"


def test_theme_can_be_moved_under_other_theme(
    admin_client, edit_link, theme, default_theme
):
    admin_client.post(edit_link, {"name": theme.name, "parent": default_theme.pk})
    theme.refresh_from_db()
    assert theme.parent == default_theme


def test_moving_theme_under_other_theme_updates_themes_tree(
    admin_client, edit_link, theme, default_theme
):
    admin_client.post(edit_link, {"name": theme.name, "parent": default_theme.pk})

    default_theme.refresh_from_db()
    theme.refresh_from_db()

    assert theme.tree_id == default_theme.tree_id
    assert theme.lft == 2
    assert theme.rght == 3

    assert default_theme.lft == 1
    assert default_theme.rght == 4


def test_theme_cant_be_moved_under_itself(admin_client, edit_link, theme):
    admin_client.post(edit_link, {"name": theme.name, "parent": theme.pk})
    theme.refresh_from_db()
    assert not theme.parent


def test_theme_cant_be_moved_under_its_child_theme(admin_client, edit_link, theme):
    child_theme = Theme.objects.create(name="Child Theme", parent=theme)
    admin_client.post(edit_link, {"name": theme.name, "parent": child_theme.pk})
    theme.refresh_from_db()
    assert not theme.parent


def test_moving_child_theme_under_other_theme_updates_both_themes_trees(
    admin_client, theme, default_theme
):
    child_theme = Theme.objects.create(name="Child Theme", parent=theme)
    edit_link = reverse("misago:admin:themes:edit", kwargs={"pk": child_theme.pk})

    admin_client.post(edit_link, {"name": child_theme.name, "parent": default_theme.pk})

    default_theme.refresh_from_db()
    child_theme.refresh_from_db()
    theme.refresh_from_db()

    assert child_theme.parent == default_theme
    assert child_theme.tree_id == default_theme.tree_id
    assert child_theme.lft == 2
    assert child_theme.rght == 3

    assert default_theme.lft == 1
    assert default_theme.rght == 4

    assert theme.tree_id != default_theme.tree_id
    assert theme.lft == 1
    assert theme.rght == 2


def test_moving_theme_under_root_updates_theme_tree_and_creates_new_one(
    admin_client, edit_link, theme, default_theme
):
    theme.parent = default_theme
    theme.save()

    admin_client.post(edit_link, {"name": theme.name})

    default_theme.refresh_from_db()
    theme.refresh_from_db()

    assert theme.parent is None
    assert theme.tree_id != default_theme.tree_id
    assert theme.lft == 1
    assert theme.rght == 2

    assert default_theme.lft == 1
    assert default_theme.rght == 2


def test_moving_theme_with_child_under_root_updates_theme_tree_and_creates_new_one(
    admin_client, edit_link, theme, default_theme
):
    theme.parent = default_theme
    theme.save()

    child_theme = Theme.objects.create(name="Child Theme", parent=theme)

    admin_client.post(edit_link, {"name": theme.name})

    default_theme.refresh_from_db()
    child_theme.refresh_from_db()
    theme.refresh_from_db()

    assert theme.tree_id != default_theme.tree_id
    assert theme.lft == 1
    assert theme.rght == 4

    assert child_theme.parent == theme
    assert child_theme.tree_id == theme.tree_id
    assert child_theme.lft == 2
    assert child_theme.rght == 3

    assert default_theme.lft == 1
    assert default_theme.rght == 2


def test_theme_edition_fails_if_parent_theme_doesnt_exist(
    admin_client, edit_link, theme, nonexisting_theme
):
    admin_client.post(
        edit_link, {"name": "Edited Theme", "parent": nonexisting_theme.pk}
    )
    theme.refresh_from_db()
    assert theme.name != "Edited Theme"


def test_error_message_is_set_if_user_attempts_to_edit_default_theme(
    admin_client, default_theme
):
    edit_link = reverse("misago:admin:themes:edit", kwargs={"pk": default_theme.pk})
    response = admin_client.get(edit_link)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_edit_nonexisting_theme(
    admin_client, nonexisting_theme
):
    edit_link = reverse("misago:admin:themes:edit", kwargs={"pk": nonexisting_theme.pk})
    response = admin_client.get(edit_link)
    assert_has_error_message(response)


def test_moving_theme_invalidates_themes_cache(
    admin_client, edit_link, theme, default_theme
):
    with assert_invalidates_cache(THEME_CACHE):
        admin_client.post(edit_link, {"name": theme.name, "parent": default_theme.pk})



import pytest

from ....notifications.models import Notification


@pytest.fixture
def notifications(user):
    objects = Notification.objects.bulk_create(
        [Notification(user=user, verb=f"test_{i}") for i in range(15)]
    )

    return sorted([obj.id for obj in objects])



from unittest.mock import Mock

from ..categories_map import get_categories_map


def test_categories_map_returns_top_categories_visible_by_user(
    cache_versions,
    user,
    default_category,
    sibling_category,
    child_category,
    other_category,
):
    request = Mock(
        cache_versions=cache_versions,
        user=user,
        user_acl={
            "visible_categories": [
                default_category.id,
                sibling_category.id,
                child_category.id,
                other_category.id,
            ],
        },
    )

    categories_map = get_categories_map(request)
    assert categories_map == [
        {
            "id": default_category.id,
            "name": default_category.name,
            "shortName": default_category.short_name,
            "color": default_category.color,
            "url": default_category.get_absolute_url(),
        },
        {
            "id": sibling_category.id,
            "name": sibling_category.name,
            "shortName": sibling_category.short_name,
            "color": sibling_category.color,
            "url": sibling_category.get_absolute_url(),
        },
        {
            "id": other_category.id,
            "name": other_category.name,
            "shortName": other_category.short_name,
            "color": other_category.color,
            "url": other_category.get_absolute_url(),
        },
    ]


def test_categories_map_excludes_categories_invisible_to_user(
    cache_versions,
    user,
    default_category,
    sibling_category,
    child_category,
    other_category,
):
    request = Mock(
        cache_versions=cache_versions,
        user=user,
        user_acl={
            "visible_categories": [
                default_category.id,
                other_category.id,
            ],
        },
    )

    categories_map = get_categories_map(request)
    assert categories_map == [
        {
            "id": default_category.id,
            "name": default_category.name,
            "shortName": default_category.short_name,
            "color": default_category.color,
            "url": default_category.get_absolute_url(),
        },
        {
            "id": other_category.id,
            "name": other_category.name,
            "shortName": other_category.short_name,
            "color": other_category.color,
            "url": other_category.get_absolute_url(),
        },
    ]


def test_categories_map_excludes_special_categories(
    cache_versions,
    user,
    private_threads_category,
    default_category,
    sibling_category,
    child_category,
    other_category,
):
    request = Mock(
        cache_versions=cache_versions,
        user=user,
        user_acl={
            "visible_categories": [
                private_threads_category.id,
                default_category.id,
                other_category.id,
            ],
        },
    )

    categories_map = get_categories_map(request)
    assert categories_map == [
        {
            "id": default_category.id,
            "name": default_category.name,
            "shortName": default_category.short_name,
            "color": default_category.color,
            "url": default_category.get_absolute_url(),
        },
        {
            "id": other_category.id,
            "name": other_category.name,
            "shortName": other_category.short_name,
            "color": other_category.color,
            "url": other_category.get_absolute_url(),
        },
    ]


def test_categories_map_skips_database_read_if_there_is_cache(
    mocker,
    django_assert_num_queries,
    cache_versions,
    user,
    default_category,
    sibling_category,
    child_category,
    other_category,
):
    cache_get = mocker.patch(
        "django.core.cache.cache.get",
        return_value=[
            {
                "id": default_category.id,
                "name": default_category.name,
                "shortName": default_category.short_name,
                "color": default_category.color,
                "url": default_category.get_absolute_url(),
            },
        ],
    )

    request = Mock(
        cache_versions=cache_versions,
        user=user,
        user_acl={
            "visible_categories": [
                default_category.id,
                other_category.id,
            ],
        },
    )

    with django_assert_num_queries(0):
        categories_map = get_categories_map(request)

    assert categories_map == [
        {
            "id": default_category.id,
            "name": default_category.name,
            "shortName": default_category.short_name,
            "color": default_category.color,
            "url": default_category.get_absolute_url(),
        },
    ]

    cache_get.assert_called_once()


def test_categories_map_is_cached_when_database_is_used(
    mocker,
    django_assert_num_queries,
    cache_versions,
    user,
    default_category,
    sibling_category,
    child_category,
    other_category,
):
    cache_set = mocker.patch("django.core.cache.cache.set")

    request = Mock(
        cache_versions=cache_versions,
        user=user,
        user_acl={
            "visible_categories": [
                default_category.id,
                other_category.id,
            ],
        },
    )

    with django_assert_num_queries(1):
        categories_map = get_categories_map(request)

    assert categories_map == [
        {
            "id": default_category.id,
            "name": default_category.name,
            "shortName": default_category.short_name,
            "color": default_category.color,
            "url": default_category.get_absolute_url(),
        },
        {
            "id": other_category.id,
            "name": other_category.name,
            "shortName": other_category.short_name,
            "color": other_category.color,
            "url": other_category.get_absolute_url(),
        },
    ]

    cache_set.assert_called_once()



from django.urls import reverse


def preload_threads_urls(request):
    request.frontend_context.update(
        {
            "ATTACHMENTS_API": reverse("misago:api:attachment-list"),
            "THREAD_EDITOR_API": reverse("misago:api:thread-editor"),
            "THREADS_API": reverse("misago:api:thread-list"),
            "PRIVATE_THREADS_API": reverse("misago:api:private-thread-list"),
            "PRIVATE_THREADS_URL": reverse("misago:private-threads"),
        }
    )

    return {}



# Generated by Django 1.10.5 on 2017-03-03 20:27
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("misago_categories", "0004_category_last_thread")]

    operations = [
        migrations.AddField(
            model_name="category",
            name="require_edits_approval",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="category",
            name="require_replies_approval",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="category",
            name="require_threads_approval",
            field=models.BooleanField(default=False),
        ),
    ]



from rest_framework import mixins, viewsets

from ..models import Rank
from ..serializers import RankSerializer


class RanksViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    serializer_class = RankSerializer
    queryset = Rank.objects.filter(is_tab=True).order_by("order")



from unittest.mock import ANY, Mock

import pytest
from ariadne import gql
from requests.exceptions import RequestException

from .... import __version__
from ..versioncheck import CACHE_KEY, CACHE_LENGTH, get_latest_version, resolve_version

test_query = gql("{ version { status message description } }")


def mock_requests_get(mocker, mock):
    return mocker.patch("requests.get", return_value=Mock(json=mock))


def test_version_check_query_returns_error_if_misago_version_is_unreleased(
    admin_graphql_client, mocker
):
    mocker.patch("misago.graphql.admin.versioncheck.__released__", False)
    mock_requests_get(mocker, Mock(return_value={"info": {"version": "outdated"}}))
    result = admin_graphql_client.query(test_query)
    assert result["version"]["status"] == "ERROR"


def test_version_check_query_returns_success_if_site_is_updated(
    admin_graphql_client, mocker
):
    mocker.patch("misago.graphql.admin.versioncheck.__released__", True)
    mock_requests_get(mocker, Mock(return_value={"info": {"version": __version__}}))
    result = admin_graphql_client.query(test_query)
    assert result["version"]["status"] == "SUCCESS"


def test_version_check_query_returns_error_if_site_is_outdated(
    admin_graphql_client, mocker
):
    mocker.patch("misago.graphql.admin.versioncheck.__released__", True)
    mock_requests_get(mocker, Mock(return_value={"info": {"version": "outdated"}}))
    result = admin_graphql_client.query(test_query)
    assert result["version"]["status"] == "ERROR"


def test_version_check_query_returns_warning_if_version_check_failed(
    admin_graphql_client, mocker
):
    mocker.patch("misago.graphql.admin.versioncheck.__released__", True)
    mock_requests_get(mocker, Mock(side_effect=RequestException()))
    result = admin_graphql_client.query(test_query)
    assert result["version"]["status"] == "WARNING"


def test_version_check_result_is_cached(mocker):
    mocker.patch("misago.graphql.admin.versioncheck.__released__", True)
    set_cache = mocker.patch("django.core.cache.cache.set")
    mock_requests_get(mocker, Mock(return_value={"info": {"version": "outdated"}}))
    resolve_version()
    set_cache.assert_called_with(CACHE_KEY, ANY, CACHE_LENGTH)


def test_failed_version_check_result_is_not_cached(admin_graphql_client, mocker):
    mocker.patch("misago.graphql.admin.versioncheck.__released__", True)
    set_cache = mocker.patch("django.core.cache.cache.set")
    mock_requests_get(mocker, Mock(side_effect=RequestException()))
    resolve_version()
    set_cache.assert_not_called()


def test_remote_api_is_not_called_if_version_check_cache_is_available(mocker):
    mocker.patch("misago.graphql.admin.versioncheck.__released__", True)
    mocker.patch("django.core.cache.cache.get", return_value={"status": "TEST"})
    api_mock = mock_requests_get(mocker, Mock())
    resolve_version()
    api_mock.assert_not_called()


def test_version_check_cache_is_returned_when_set(mocker):
    mocker.patch("misago.graphql.admin.versioncheck.__released__", True)
    mocker.patch("django.core.cache.cache.get", return_value={"status": "TEST"})
    mock_requests_get(mocker, Mock())
    assert get_latest_version() == {"status": "TEST"}



from ..urlpatterns import discover_plugins_urlpatterns


def test_discover_plugins_urlpatterns_discovers_apps_with_urls():
    urlpatterns = discover_plugins_urlpatterns(["misago.threads", "misago.users"])
    assert len(urlpatterns) == 2


def test_discover_plugins_urlpatterns_skips_apps_without_urls():
    urlpatterns = discover_plugins_urlpatterns(["misago.core"])
    assert not urlpatterns



from rest_framework import serializers

from ...core.utils import format_plaintext_for_html
from ..models import Rank

__all__ = ["RankSerializer"]


class RankSerializer(serializers.ModelSerializer):
    description = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = Rank
        fields = [
            "id",
            "name",
            "slug",
            "description",
            "title",
            "css_class",
            "is_default",
            "is_tab",
            "url",
        ]

    def get_name(self, obj):
        return str(obj)

    def get_description(self, obj):
        if obj.description:
            return format_plaintext_for_html(obj.description)
        return ""

    def get_url(self, obj):
        return obj.get_absolute_url()



import html
from dataclasses import dataclass

import html5lib

SINGLETON_TAGS = (
    "area",
    "base",
    "br",
    "col",
    "command",
    "embed",
    "hr",
    "img",
    "input",
    "keygen",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
)


class Node:
    def __str__(self):
        raise NotImplementedError("Subclasses of 'Node' need to implement __str__")


@dataclass
class RootNode(Node):
    tag = None
    children: list

    def __str__(self):
        return "".join(str(child) for child in self.children)


@dataclass
class ElementNode(Node):
    tag: str
    attrs: dict
    children: list

    def __str__(self):
        attrs_padding = " " if self.attrs else ""
        attrs = " ".join(self.attrs_str())

        if self.tag in SINGLETON_TAGS:
            return f"<{self.tag}{attrs_padding}{attrs} />"

        children = "".join(str(child) for child in self.children)
        return f"<{self.tag}{attrs_padding}{attrs}>{children}</{self.tag}>"

    def attrs_str(self):
        for name, value in self.attrs.items():
            if value is True or not value:
                yield html.escape(str(name))
            else:
                yield (f'{html.escape(str(name))}="{html.escape(str(value))}"')


@dataclass
class TextNode(Node):
    text: str

    def __str__(self):
        return html.escape(self.text)


def parse_html_string(string: str) -> RootNode:
    element = html5lib.parse(
        string,
        namespaceHTMLElements=False,
    )

    body = element.find("body")
    root_node = RootNode(children=[])

    if body.text:
        root_node.children.append(TextNode(text=body.text))

    for child in body:
        add_child_node(root_node, child)

    return root_node


def add_child_node(parent, element):
    node = ElementNode(
        tag=element.tag,
        attrs=element.attrib,
        children=[],
    )

    if element.text:
        node.children.append(TextNode(text=element.text))

    parent.children.append(node)

    if element.tail:
        parent.children.append(TextNode(text=element.tail))

    for child in element:
        add_child_node(node, child)


def print_html_string(root_node: RootNode) -> str:
    return str(root_node)



from ..threads import (
    get_fake_closed_thread,
    get_fake_hidden_thread,
    get_fake_thread,
    get_fake_unapproved_thread,
)


def test_fake_thread_can_be_created(fake, default_category):
    assert get_fake_thread(fake, default_category)


def test_fake_thread_is_created_with_opening_post(fake, default_category):
    thread = get_fake_thread(fake, default_category)
    assert thread.first_post


def test_fake_thread_is_created_with_guest_starter(fake, default_category):
    thread = get_fake_thread(fake, default_category)
    assert thread.first_post.poster is None


def test_fake_thread_is_created_with_specified_starter(fake, default_category, user):
    thread = get_fake_thread(fake, default_category, user)
    assert thread.first_post.poster == user
    assert thread.first_post.poster_name == user.username


def test_fake_thread_is_created_in_specified_category(fake, default_category):
    thread = get_fake_thread(fake, default_category)
    assert thread.category == default_category
    assert thread.first_post.category == default_category


def test_fake_closed_thread_can_be_created(fake, default_category):
    thread = get_fake_closed_thread(fake, default_category)
    assert thread.is_closed


def test_fake_hidden_thread_can_be_created(fake, default_category):
    thread = get_fake_hidden_thread(fake, default_category)
    assert thread.is_hidden
    assert thread.first_post.is_hidden


def test_fake_unapproved_thread_can_be_created(fake, default_category):
    thread = get_fake_unapproved_thread(fake, default_category)
    assert thread.is_unapproved
    assert thread.first_post.is_unapproved


def test_different_fake_thread_title_is_used_every_time(fake, default_category):
    thread_a = get_fake_thread(fake, default_category)
    thread_b = get_fake_thread(fake, default_category)
    assert thread_a.title != thread_b.title



from django.utils.translation import pgettext
from rest_framework import serializers

from ...core.utils import format_plaintext_for_html
from ..models import Ban

__all__ = ["BanMessageSerializer", "BanDetailsSerializer"]


def serialize_message(message):
    if message:
        return {"plain": message, "html": format_plaintext_for_html(message)}


class BanMessageSerializer(serializers.ModelSerializer):
    message = serializers.SerializerMethodField()

    class Meta:
        model = Ban
        fields = ["message", "expires_on"]

    def get_message(self, obj):
        if obj.user_message:
            message = obj.user_message
        elif obj.check_type == Ban.IP:
            message = pgettext("ban message", "Your IP address is banned.")
        else:
            message = pgettext("ban message", "You are banned.")

        return serialize_message(message)


class BanDetailsSerializer(serializers.ModelSerializer):
    user_message = serializers.SerializerMethodField()
    staff_message = serializers.SerializerMethodField()

    class Meta:
        model = Ban
        fields = ["user_message", "staff_message", "expires_on"]

    def get_user_message(self, obj):
        return serialize_message(obj.user_message)

    def get_staff_message(self, obj):
        return serialize_message(obj.staff_message)



from django.urls import include, path

from ...core.views import home_redirect

from ..views import (
    activation,
    auth,
    avatarserver,
    forgottenpassword,
    lists,
    options,
    profile,
)

urlpatterns = [
    path("banned/", home_redirect, name="banned"),
    path("login/", auth.login, name="login"),
    path("logout/", auth.logout, name="logout"),
    path(
        "request-activation/",
        activation.request_activation,
        name="request-activation",
    ),
    path(
        "activation/<int:pk>/<slug:token>/",
        activation.activate_by_token,
        name="activate-by-token",
    ),
    path(
        "forgotten-password/",
        forgottenpassword.request_reset,
        name="forgotten-password",
    ),
    path(
        "forgotten-password/<int:pk>/<slug:token>/",
        forgottenpassword.reset_password_form,
        name="forgotten-password-change-form",
    ),
]

urlpatterns += [
    path("options/", options.index, name="options"),
    path("options/<slug:form_name>/", options.index, name="options-form"),
    path("options/forum-options/", options.index, name="usercp-change-forum-options"),
    path("options/edit-details/", options.index, name="usercp-edit-details"),
    path("options/change-username/", options.index, name="usercp-change-username"),
    path(
        "options/sign-in-credentials/",
        options.index,
        name="usercp-change-email-password",
    ),
    path(
        "options/change-email/<slug:token>/",
        options.confirm_email_change,
        name="options-confirm-email-change",
    ),
    path(
        "options/change-password/<slug:token>/",
        options.confirm_password_change,
        name="options-confirm-password-change",
    ),
    path("options/download-data/", options.index, name="usercp-download-data"),
    path("options/delete-account/", options.index, name="usercp-delete-account"),
]

urlpatterns += [
    path(
        "users/",
        include(
            [
                path("", lists.landing, name="users"),
                path(
                    "active-posters/",
                    lists.ActivePostersView.as_view(),
                    name="users-active-posters",
                ),
                path(
                    "<slug:slug>/",
                    lists.RankUsersView.as_view(),
                    name="users-rank",
                ),
                path(
                    "<slug:slug>/<int:page>/",
                    lists.RankUsersView.as_view(),
                    name="users-rank",
                ),
            ]
        ),
    )
]

urlpatterns += [
    path(
        "u/<slug:slug>/<int:pk>/",
        include(
            [
                path("", profile.LandingView.as_view(), name="user"),
                path("posts/", profile.UserPostsView.as_view(), name="user-posts"),
                path(
                    "threads/",
                    profile.UserThreadsView.as_view(),
                    name="user-threads",
                ),
                path(
                    "followers/",
                    profile.UserFollowersView.as_view(),
                    name="user-followers",
                ),
                path(
                    "follows/",
                    profile.UserFollowsView.as_view(),
                    name="user-follows",
                ),
                path(
                    "details/",
                    profile.UserProfileDetailsView.as_view(),
                    name="user-details",
                ),
                path(
                    "username-history/",
                    profile.UserUsernameHistoryView.as_view(),
                    name="username-history",
                ),
                path("ban-details/", profile.UserBanView.as_view(), name="user-ban"),
            ]
        ),
    )
]

urlpatterns += [
    path("avatar/", avatarserver.blank_avatar, name="blank-avatar"),
    path(
        "avatar/<int:pk>/<int:size>/",
        avatarserver.user_avatar,
        name="user-avatar",
    ),
]



# Generated by Django 1.11.16 on 2018-11-25 15:31
from django.db import migrations

from ...cache.operations import StartCacheVersioning


class Migration(migrations.Migration):
    dependencies = [
        ("misago_users", "0015_user_agreements"),
        ("misago_cache", "0001_initial"),
    ]

    operations = [StartCacheVersioning("bans")]



from django.urls import reverse

from ...acl.test import patch_user_acl
from ...threads.search import SearchThreads
from ...users.test import AuthenticatedUserTestCase


class LandingTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.test_link = reverse("misago:search")

    @patch_user_acl({"can_search": False})
    def test_no_permission(self):
        """view validates permission to search forum"""
        response = self.client.get(self.test_link)
        self.assertContains(response, "have permission to search site", status_code=403)

    @patch_user_acl({"can_search": True})
    def test_redirect_to_provider(self):
        """view validates permission to search forum"""
        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 302)
        self.assertIn(SearchThreads.url, response["location"])


class SearchTests(AuthenticatedUserTestCase):
    @patch_user_acl({"can_search": False})
    def test_no_permission(self):
        """view validates permission to search forum"""
        response = self.client.get(
            reverse("misago:search", kwargs={"search_provider": "users"})
        )

        self.assertContains(response, "have permission to search site", status_code=403)

    def test_not_found(self):
        """view raises 404 for not found provider"""
        response = self.client.get(
            reverse("misago:search", kwargs={"search_provider": "nada"})
        )

        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_search": True, "can_search_users": False})
    def test_provider_no_permission(self):
        """provider raises 403 without permission"""
        response = self.client.get(
            reverse("misago:search", kwargs={"search_provider": "users"})
        )

        self.assertContains(
            response, "have permission to search users", status_code=403
        )

    def test_provider(self):
        """provider displays no script page"""
        response = self.client.get(
            reverse("misago:search", kwargs={"search_provider": "users"})
        )

        self.assertContains(response, "Loading search...")



from django.test import TestCase
from django.urls import reverse

from ...conf import settings
from ..test import create_test_user


class AvatarServerTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "User@example.com")

    def test_get_user_avatar_exact_size(self):
        """avatar server resolved valid avatar url for user"""
        avatar_url = reverse(
            "misago:user-avatar", kwargs={"pk": self.user.pk, "size": 200}
        )

        response = self.client.get(avatar_url)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], self.user.avatars[1]["url"])

    def test_get_user_avatar_inexact_size(self):
        """avatar server resolved valid avatar fallback for user"""
        avatar_url = reverse(
            "misago:user-avatar", kwargs={"pk": self.user.pk, "size": 250}
        )

        response = self.client.get(avatar_url)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], self.user.avatars[0]["url"])

    def test_get_notfound_user_avatar(self):
        """avatar server handles deleted user avatar requests"""
        avatar_url = reverse(
            "misago:user-avatar", kwargs={"pk": self.user.pk + 1, "size": 150}
        )
        response = self.client.get(avatar_url)

        self.assertEqual(response.status_code, 302)
        self.assertTrue(response["location"].endswith(settings.MISAGO_BLANK_AVATAR))

    def test_blank_avatar_serving(self):
        """avatar server handles blank avatar requests"""
        response = self.client.get(reverse("misago:blank-avatar"))

        self.assertEqual(response.status_code, 302)
        self.assertTrue(response["location"].endswith(settings.MISAGO_BLANK_AVATAR))



from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase


class PrivateThreadsTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.category = Category.objects.private_threads()



from django.urls import path

from .. import admin
from .views import auth, index

urlpatterns = [
    # "misago:admin:index" link symbolises "root" of Misago admin links space
    # any request with path that falls below this one is assumed to be directed
    # at Misago Admin and will be checked by Misago Admin Middleware
    path("", index.admin_index, name="index"),
    path("logout/", auth.logout, name="logout"),
]

# Discover admin and register patterns
admin.discover_misago_admin()
urlpatterns += admin.urlpatterns()



from datetime import timedelta

from django.test import TestCase
from django.utils import timezone

from ...conftest import get_cache_versions
from ..bans import (
    ban_ip,
    ban_user,
    get_email_ban,
    get_ip_ban,
    get_request_ip_ban,
    get_user_ban,
    get_username_ban,
)
from ..models import Ban
from ..test import create_test_user

cache_versions = get_cache_versions()


class GetBanTests(TestCase):
    def test_get_username_ban(self):
        """get_username_ban returns valid ban"""
        nonexistent_ban = get_username_ban("nonexistent")
        self.assertIsNone(nonexistent_ban)

        Ban.objects.create(
            banned_value="expired", expires_on=timezone.now() - timedelta(days=7)
        )

        expired_ban = get_username_ban("expired")
        self.assertIsNone(expired_ban)

        Ban.objects.create(banned_value="wrongtype", check_type=Ban.EMAIL)

        wrong_type_ban = get_username_ban("wrongtype")
        self.assertIsNone(wrong_type_ban)

        valid_ban = Ban.objects.create(
            banned_value="admi*", expires_on=timezone.now() + timedelta(days=7)
        )
        self.assertEqual(get_username_ban("admiral").pk, valid_ban.pk)

        registration_ban = Ban.objects.create(
            banned_value="mod*",
            expires_on=timezone.now() + timedelta(days=7),
            registration_only=True,
        )
        self.assertIsNone(get_username_ban("moderator"))
        self.assertEqual(get_username_ban("moderator", True).pk, registration_ban.pk)

    def test_get_email_ban(self):
        """get_email_ban returns valid ban"""
        nonexistent_ban = get_email_ban("non@existent.com")
        self.assertIsNone(nonexistent_ban)

        Ban.objects.create(
            banned_value="ex@pired.com",
            check_type=Ban.EMAIL,
            expires_on=timezone.now() - timedelta(days=7),
        )

        expired_ban = get_email_ban("ex@pired.com")
        self.assertIsNone(expired_ban)

        Ban.objects.create(banned_value="wrong@type.com", check_type=Ban.IP)

        wrong_type_ban = get_email_ban("wrong@type.com")
        self.assertIsNone(wrong_type_ban)

        valid_ban = Ban.objects.create(
            banned_value="*.ru",
            check_type=Ban.EMAIL,
            expires_on=timezone.now() + timedelta(days=7),
        )
        self.assertEqual(get_email_ban("banned@mail.ru").pk, valid_ban.pk)

        registration_ban = Ban.objects.create(
            banned_value="*.ua",
            check_type=Ban.EMAIL,
            expires_on=timezone.now() + timedelta(days=7),
            registration_only=True,
        )
        self.assertIsNone(get_email_ban("banned@mail.ua"))
        self.assertEqual(get_email_ban("banned@mail.ua", True).pk, registration_ban.pk)

    def test_get_ip_ban(self):
        """get_ip_ban returns valid ban"""
        nonexistent_ban = get_ip_ban("123.0.0.1")
        self.assertIsNone(nonexistent_ban)

        Ban.objects.create(
            banned_value="124.0.0.1",
            check_type=Ban.IP,
            expires_on=timezone.now() - timedelta(days=7),
        )

        expired_ban = get_ip_ban("124.0.0.1")
        self.assertIsNone(expired_ban)

        Ban.objects.create(banned_value="wrongtype", check_type=Ban.EMAIL)

        wrong_type_ban = get_ip_ban("wrongtype")
        self.assertIsNone(wrong_type_ban)

        valid_ban = Ban.objects.create(
            banned_value="125.0.0.*",
            check_type=Ban.IP,
            expires_on=timezone.now() + timedelta(days=7),
        )
        self.assertEqual(get_ip_ban("125.0.0.1").pk, valid_ban.pk)

        registration_ban = Ban.objects.create(
            banned_value="188.*",
            check_type=Ban.IP,
            expires_on=timezone.now() + timedelta(days=7),
            registration_only=True,
        )
        self.assertIsNone(get_ip_ban("188.12.12.41"))
        self.assertEqual(get_ip_ban("188.12.12.41", True).pk, registration_ban.pk)


class UserBansTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com")

    def test_no_ban(self):
        """user is not caught by ban"""
        self.assertIsNone(get_user_ban(self.user, cache_versions))
        self.assertFalse(self.user.ban_cache.is_banned)

    def test_permanent_ban(self):
        """user is caught by permanent ban"""
        Ban.objects.create(
            banned_value="User",
            user_message="User reason",
            staff_message="Staff reason",
        )

        user_ban = get_user_ban(self.user, cache_versions)
        self.assertIsNotNone(user_ban)
        self.assertEqual(user_ban.user_message, "User reason")
        self.assertEqual(user_ban.staff_message, "Staff reason")
        self.assertTrue(self.user.ban_cache.is_banned)

    def test_temporary_ban(self):
        """user is caught by temporary ban"""
        Ban.objects.create(
            banned_value="us*",
            user_message="User reason",
            staff_message="Staff reason",
            expires_on=timezone.now() + timedelta(days=7),
        )

        user_ban = get_user_ban(self.user, cache_versions)
        self.assertIsNotNone(user_ban)
        self.assertEqual(user_ban.user_message, "User reason")
        self.assertEqual(user_ban.staff_message, "Staff reason")
        self.assertTrue(self.user.ban_cache.is_banned)

    def test_expired_ban(self):
        """user is not caught by expired ban"""
        Ban.objects.create(
            banned_value="us*", expires_on=timezone.now() - timedelta(days=7)
        )

        self.assertIsNone(get_user_ban(self.user, cache_versions))
        self.assertFalse(self.user.ban_cache.is_banned)

    def test_expired_non_flagged_ban(self):
        """user is not caught by expired but checked ban"""
        Ban.objects.create(
            banned_value="us*", expires_on=timezone.now() - timedelta(days=7)
        )
        Ban.objects.update(is_checked=True)

        self.assertIsNone(get_user_ban(self.user, cache_versions))
        self.assertFalse(self.user.ban_cache.is_banned)


class MockRequest:
    def __init__(self):
        self.user_ip = "127.0.0.1"
        self.session = {}
        self.cache_versions = cache_versions


class RequestIPBansTests(TestCase):
    def test_no_ban(self):
        """no ban found"""
        ip_ban = get_request_ip_ban(MockRequest())
        self.assertIsNone(ip_ban)

    def test_permanent_ban(self):
        """ip is caught by permanent ban"""
        Ban.objects.create(
            check_type=Ban.IP, banned_value="127.0.0.1", user_message="User reason"
        )

        ip_ban = get_request_ip_ban(MockRequest())
        self.assertTrue(ip_ban["is_banned"])
        self.assertEqual(ip_ban["ip"], "127.0.0.1")
        self.assertEqual(ip_ban["message"], "User reason")

        # repeated call uses cache
        get_request_ip_ban(MockRequest())

    def test_temporary_ban(self):
        """ip is caught by temporary ban"""
        Ban.objects.create(
            check_type=Ban.IP,
            banned_value="127.0.0.1",
            user_message="User reason",
            expires_on=timezone.now() + timedelta(days=7),
        )

        ip_ban = get_request_ip_ban(MockRequest())
        self.assertTrue(ip_ban["is_banned"])
        self.assertEqual(ip_ban["ip"], "127.0.0.1")
        self.assertEqual(ip_ban["message"], "User reason")

        # repeated call uses cache
        get_request_ip_ban(MockRequest())

    def test_expired_ban(self):
        """ip is not caught by expired ban"""
        Ban.objects.create(
            check_type=Ban.IP,
            banned_value="127.0.0.1",
            user_message="User reason",
            expires_on=timezone.now() - timedelta(days=7),
        )

        ip_ban = get_request_ip_ban(MockRequest())
        self.assertIsNone(ip_ban)

        # repeated call uses cache
        get_request_ip_ban(MockRequest())


class BanUserTests(TestCase):
    def test_ban_user(self):
        """ban_user utility bans user"""
        user = create_test_user("User", "user@example.com")

        ban = ban_user(user, "User reason", "Staff reason")
        self.assertEqual(ban.user_message, "User reason")
        self.assertEqual(ban.staff_message, "Staff reason")

        db_ban = get_user_ban(user, cache_versions)
        self.assertEqual(ban.pk, db_ban.ban_id)


class BanIpTests(TestCase):
    def test_ban_ip(self):
        """ban_ip utility bans IP address"""
        ban = ban_ip("127.0.0.1", "User reason", "Staff reason")
        self.assertEqual(ban.user_message, "User reason")
        self.assertEqual(ban.staff_message, "Staff reason")

        db_ban = get_ip_ban("127.0.0.1")
        self.assertEqual(ban.pk, db_ban.pk)



import responses
from django.contrib.auth import get_user_model
from django.urls import reverse
from responses.matchers import header_matcher, urlencoded_params_matcher

from ...conf.test import override_dynamic_settings
from ...test import assert_contains
from ...users.bans import ban_ip, ban_user
from ..client import SESSION_STATE
from ..models import Subject

User = get_user_model()


def test_oauth2_complete_view_returns_404_if_oauth_is_disabled(
    client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is False

    response = client.get(reverse("misago:oauth2-complete"))
    assert response.status_code == 404


def test_oauth2_complete_view_returns_error_404_if_user_ip_is_banned_and_oauth_is_disabled(
    client, dynamic_settings
):
    ban_ip("127.*", "Ya got banned!")

    assert dynamic_settings.enable_oauth2_client is False

    response = client.get(reverse("misago:oauth2-complete"))
    assert response.status_code == 404


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_client_id="clientid123",
    oauth2_scopes="scopes",
    oauth2_login_url="https://example.com/oauth2/login",
)
def test_oauth2_complete_view_returns_error_403_if_user_ip_is_banned(
    client, dynamic_settings
):
    ban_ip("127.*", "Ya got banned!")

    assert dynamic_settings.enable_oauth2_client is True

    response = client.get(reverse("misago:oauth2-complete"))
    assert_contains(response, "Ya got banned", 403)


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_client_id="clientid123",
    oauth2_scopes="scopes",
    oauth2_login_url="https://example.com/oauth2/login",
)
def test_oauth2_complete_view_returns_error_400_if_user_canceled_sign_in(
    client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    response = client.get("%s?error=access_denied" % reverse("misago:oauth2-complete"))
    assert_contains(response, "The OAuth2 process was canceled by the provider.", 400)


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_client_id="clientid123",
    oauth2_scopes="scopes",
    oauth2_login_url="https://example.com/oauth2/login",
)
def test_oauth2_complete_view_returns_error_400_if_state_is_not_set(
    client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    response = client.get(reverse("misago:oauth2-complete"))
    assert_contains(response, "OAuth2 session is missing state.", 400)


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_client_id="clientid123",
    oauth2_scopes="scopes",
    oauth2_login_url="https://example.com/oauth2/login",
)
def test_oauth2_complete_view_returns_error_400_if_state_is_invalid(
    client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    session = client.session
    session[SESSION_STATE] = "state123"
    session.save()

    response = client.get(
        "%s?state=invalid&code=1234" % reverse("misago:oauth2-complete")
    )
    assert_contains(
        response,
        "OAuth2 state sent by the provider did not match one in the session.",
        400,
    )


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_client_id="clientid123",
    oauth2_scopes="scopes",
    oauth2_login_url="https://example.com/oauth2/login",
)
def test_oauth2_complete_view_returns_error_400_if_code_is_missing(
    client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    session = client.session
    session[SESSION_STATE] = "state123"
    session.save()

    response = client.get("%s?state=state123&code=" % reverse("misago:oauth2-complete"))
    assert_contains(
        response,
        "OAuth2 authorization code was not sent by the provider.",
        400,
    )


TEST_SETTINGS = {
    "enable_oauth2_client": True,
    "oauth2_client_id": "oauth2_client_id",
    "oauth2_client_secret": "oauth2_client_secret",
    "oauth2_login_url": "https://example.com/oauth2/login",
    "oauth2_token_url": "https://example.com/oauth2/token",
    "oauth2_json_token_path": "token.bearer",
    "oauth2_user_url": "https://example.com/oauth2/user",
    "oauth2_user_method": "POST",
    "oauth2_user_token_name": "Authorization",
    "oauth2_user_token_location": "HEADER_BEARER",
    "oauth2_send_welcome_email": True,
    "oauth2_json_id_path": "id",
    "oauth2_json_name_path": "profile.name",
    "oauth2_json_email_path": "profile.email",
}


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_creates_new_user(client, dynamic_settings, mailoutbox):
    assert dynamic_settings.enable_oauth2_client is True

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "john@example.com",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert response.status_code == 302

    # User and subject are created
    subject = Subject.objects.get(sub="1234")
    user = User.objects.get_by_email("john@example.com")
    assert subject.user_id == user.id

    assert user.username == "John_Doe"
    assert user.slug == "john-doe"
    assert user.email == "john@example.com"
    assert user.requires_activation == User.ACTIVATION_NONE

    # User is authenticated
    auth_api = client.get(reverse("misago:api:auth")).json()
    assert auth_api["id"] == user.id

    # User welcome e-mail is sent
    assert len(mailoutbox) == 1
    assert mailoutbox[0].subject == "Welcome on Misago forums!"


TEST_SETTINGS_EMAIL_DISABLED = TEST_SETTINGS.copy()
TEST_SETTINGS_EMAIL_DISABLED["oauth2_send_welcome_email"] = False


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS_EMAIL_DISABLED)
def test_oauth2_complete_view_doesnt_send_welcome_mail_if_option_is_disabled(
    client, dynamic_settings, mailoutbox
):
    assert dynamic_settings.enable_oauth2_client is True

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "john@example.com",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert response.status_code == 302

    # User and subject are created
    subject = Subject.objects.get(sub="1234")
    user = User.objects.get_by_email("john@example.com")
    assert subject.user_id == user.id

    # User is authenticated
    auth_api = client.get(reverse("misago:api:auth")).json()
    assert auth_api["id"] == user.id

    # User welcome e-mail is not sent
    assert len(mailoutbox) == 0


TEST_SETTINGS_EXTRA_TOKEN_HEADERS = TEST_SETTINGS.copy()
TEST_SETTINGS_EXTRA_TOKEN_HEADERS[
    "oauth2_token_extra_headers"
] = """
Accept: application/json
API-Version: 2.1.3.7
""".strip()


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS_EXTRA_TOKEN_HEADERS)
def test_oauth2_complete_view_includes_extra_headers_in_token_request(
    client, dynamic_settings, mailoutbox
):
    assert dynamic_settings.enable_oauth2_client is True

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
            header_matcher(
                {
                    "Accept": "application/json",
                    "API-Version": "2.1.3.7",
                }
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "john@example.com",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert response.status_code == 302

    # User and subject are created
    subject = Subject.objects.get(sub="1234")
    user = User.objects.get_by_email("john@example.com")
    assert subject.user_id == user.id

    # User is authenticated
    auth_api = client.get(reverse("misago:api:auth")).json()
    assert auth_api["id"] == user.id


TEST_SETTINGS_EXTRA_USER_HEADERS = TEST_SETTINGS.copy()
TEST_SETTINGS_EXTRA_USER_HEADERS[
    "oauth2_user_extra_headers"
] = """
X-Header: its-a-test
API-Version: 2.1.3.7
""".strip()


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS_EXTRA_USER_HEADERS)
def test_oauth2_complete_view_includes_extra_headers_in_user_request(
    client, dynamic_settings, mailoutbox
):
    assert dynamic_settings.enable_oauth2_client is True

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "john@example.com",
            },
        },
        match=[
            header_matcher(
                {
                    "Authorization": f"Bearer {access_token}",
                    "X-Header": "its-a-test",
                    "API-Version": "2.1.3.7",
                }
            ),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert response.status_code == 302

    # User and subject are created
    subject = Subject.objects.get(sub="1234")
    user = User.objects.get_by_email("john@example.com")
    assert subject.user_id == user.id

    # User is authenticated
    auth_api = client.get(reverse("misago:api:auth")).json()
    assert auth_api["id"] == user.id


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_updates_existing_user(
    user, client, dynamic_settings, mailoutbox
):
    assert dynamic_settings.enable_oauth2_client is True

    Subject.objects.create(sub="1234", user=user)

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "john@example.com",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert response.status_code == 302

    # User is updated
    user.refresh_from_db()
    assert user.username == "John_Doe"
    assert user.slug == "john-doe"
    assert user.email == "john@example.com"

    # User is authenticated
    auth_api = client.get(reverse("misago:api:auth")).json()
    assert auth_api["id"] == user.id

    # User welcome e-mail is not sent
    assert len(mailoutbox) == 0


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_returns_error_400_if_code_grant_is_rejected(
    client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    code_grant = "12345grant"
    session_state = "12345state"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "error": "Permission denied",
        },
        status=403,
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert_contains(
        response,
        "The OAuth2 provider responded with error for an access token request.",
        400,
    )


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_returns_error_400_if_access_token_is_rejected(
    user, client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    Subject.objects.create(sub="1234", user=user)

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "error": "Permission denied",
        },
        status=403,
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert_contains(
        response,
        "The OAuth2 provider responded with error for user profile request.",
        400,
    )


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_returns_error_400_if_user_email_was_missing(
    user, client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    Subject.objects.create(sub="1234", user=user)

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert_contains(response, "Enter a valid e-mail address.", 400)


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_returns_error_400_if_user_email_was_invalid(
    user, client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    Subject.objects.create(sub="1234", user=user)

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "invalid",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert_contains(response, "Enter a valid e-mail address.", 400)


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_returns_error_400_if_user_data_causes_integrity_error(
    user, client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": user.email,
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert_contains(
        response,
        (
            "Your e-mail address returned by the provider is not available "
            "for use on this site."
        ),
        400,
    )


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_updates_deactivated_user_but_returns_error_400(
    user, client, dynamic_settings, mailoutbox
):
    assert dynamic_settings.enable_oauth2_client is True

    Subject.objects.create(sub="1234", user=user)

    user.is_active = False
    user.save()

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "john@example.com",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert_contains(
        response,
        (
            "User account associated with the profile from the OAuth2 provider "
            "was deactivated by the site administrator."
        ),
        400,
    )

    # User is updated but still deactivated
    user.refresh_from_db()
    assert user.username == "John_Doe"
    assert user.slug == "john-doe"
    assert user.email == "john@example.com"
    assert user.is_active is False

    # User is not authenticated
    auth_api = client.get(reverse("misago:api:auth")).json()
    assert auth_api["id"] is None

    # User welcome e-mail is not sent
    assert len(mailoutbox) == 0


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_creates_banned_user_but_returns_error_403(
    user, client, dynamic_settings, mailoutbox
):
    assert dynamic_settings.enable_oauth2_client is True

    user.username = "John_Doe"
    ban_user(user, "Banned for a test.")
    user.refresh_from_db()

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "john@example.com",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert_contains(response, "Banned for a test.", 403)

    # User is created
    new_user = User.objects.get_by_email("john@example.com")
    assert new_user
    assert new_user.id != user.id
    assert new_user.username == "John_Doe"
    assert new_user.slug == "john-doe"
    assert new_user.email == "john@example.com"

    # User is not authenticated
    auth_api = client.get(reverse("misago:api:auth")).json()
    assert auth_api["id"] is None

    # User welcome e-mail is not sent
    assert len(mailoutbox) == 0


@responses.activate
@override_dynamic_settings(**TEST_SETTINGS)
def test_oauth2_complete_view_updates_banned_user_but_returns_error_403(
    user, client, dynamic_settings, mailoutbox
):
    assert dynamic_settings.enable_oauth2_client is True

    Subject.objects.create(sub="1234", user=user)

    user.username = "John_Doe"
    ban_user(user, "Banned for a test.")
    user.refresh_from_db()

    code_grant = "12345grant"
    session_state = "12345state"
    access_token = "12345token"

    session = client.session
    session[SESSION_STATE] = session_state
    session.save()

    responses.post(
        "https://example.com/oauth2/token",
        json={
            "token": {
                "bearer": access_token,
            },
        },
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "authorization_code",
                    "client_id": "oauth2_client_id",
                    "client_secret": "oauth2_client_secret",
                    "redirect_uri": "http://testserver/oauth2/complete/",
                    "code": code_grant,
                },
            ),
        ],
    )

    responses.post(
        "https://example.com/oauth2/user",
        json={
            "id": 1234,
            "profile": {
                "name": "John Doe",
                "email": "john@example.com",
            },
        },
        match=[
            header_matcher({"Authorization": f"Bearer {access_token}"}),
        ],
    )

    response = client.get(
        "%s?state=%s&code=%s"
        % (
            reverse("misago:oauth2-complete"),
            session_state,
            code_grant,
        )
    )

    assert_contains(response, "Banned for a test.", 403)

    # User is updated
    user.refresh_from_db()
    assert user.username == "John_Doe"
    assert user.slug == "john-doe"
    assert user.email == "john@example.com"

    # User is not authenticated
    auth_api = client.get(reverse("misago:api:auth")).json()
    assert auth_api["id"] is None

    # User welcome e-mail is not sent
    assert len(mailoutbox) == 0



from io import BytesIO

import requests
from PIL import Image

from . import store


class AvatarDownloadError(RuntimeError):
    pass


def set_avatar(user, avatar_url):
    try:
        r = requests.get(avatar_url, timeout=5)
        if r.status_code != 200:
            raise AvatarDownloadError("could not retrieve avatar from the server")

        image = Image.open(BytesIO(r.content))

        # Crop avatar into square
        width, height = image.size

        if width > height:
            left = int((width - height) / 2)
            image = image.crop((left, 0, width + left, height))
        elif height > width:
            top = int((height - width) / 2)
            image = image.crop((0, top, width, top + height))

        store.store_new_avatar(user, image)
    except requests.exceptions.RequestException:
        raise AvatarDownloadError("failed to connect to avatar server")
